
/*
 This file is part of moth. moth is a program for creating
 and editing textured 3D models.
 
 Copyright (C) 2004 Peter Uray.

 moth is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 moth is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with moth; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdlib.h>
#include <iostream>

#include <MMainWindow.h>

#include <moth/MSceneGraph.h>
#include <moth/MItemReader.h>

#include <MPreferences.h>
#include <MDocument.h>

MDocument *MDocument::_this = NULL;

MDocument::MDocument(void)
{
 _past.setAutoDelete(true);
 _future.setAutoDelete(true);

 _selected = NULL;
 _master = NULL;
  
 _lock = false;
}


MDocument::~MDocument(void)
{

}


void MDocument::clear(void)
{ 
 MSceneGraph::get()->clear();
 _past.clear();
 _future.clear();
 
 setSelectedItem(NULL);
}


void MDocument::setSelectedItem(MSceneGraphItem *item)
{
 _selected = item;
}


MSceneGraphItem *MDocument::getSelectedItem()
{
 return _selected;
}


void MDocument::setMaster(void)
{
 if(isShapePtr(_selected))
 {
  _master = MShapePtr(_selected);
 }
 else
 {
  _master = NULL;
 }
}


void MDocument::clearMaster(void)
{
_master = NULL;
}



MShape *MDocument::getMaster(void)
{
 return _master;
}



MModel *MDocument::getSelectedModel(void)
{
 if(!_selected) return NULL;
 
 if(_selected->getType() == MEnum::Model) return (MModel *)_selected;
 return NULL;
}


MShape *MDocument::getSelectedShape(void)
{
 if(!_selected) return NULL;
 
 if(_selected->getType() == MEnum::Shape) return (MShape *)_selected;
 return NULL;
}


MLightSource *MDocument::getSelectedLightSource(void)
{
 if(!_selected) return NULL;
 
 if(_selected->getType() == MEnum::LightSource) return (MLightSource *)_selected;
 return NULL;
}




void MDocument::commit(void)
{
 if(_lock) return;

 if(canRedo())
 {
  while(_future.count() > 0)
  {
   _future.first();
   _past.prepend(_future.take());
  }
 }

 push(_past, new QString(MSceneGraph::get()->toString()));
 MMainWindow::get()->setHistoryButtonState(canUndo(), canRedo());
 
 while((int)_past.count() > MPreferences::get()->getUndoSteps() || _past.count() == 0)
 {
  _past.removeLast();
 }
}


void MDocument::undo(void)
{
 if(canUndo())
 {
  push(_future, new QString(MSceneGraph::get()->toString()));
  
  MSceneGraph::get()->clear();
  MItemReader::get()->realize(*(pop(_past)));
  
  _master = NULL;
  _selected = NULL;
 }
 
 MMainWindow::get()->setHistoryButtonState(canUndo(), canRedo());
}


void MDocument::redo(void)
{
 if(canRedo())
 {
  push(_past, new QString(MSceneGraph::get()->toString()));

  MSceneGraph::get()->clear();
  MItemReader::get()->realize(*(pop(_future)));
  
  _master = NULL;
  _selected = NULL;
 }

 MMainWindow::get()->setHistoryButtonState(canUndo(), canRedo());
}


bool MDocument::canUndo(void)
{
 return !_past.isEmpty();
}


bool MDocument::canRedo(void)
{
 return !_future.isEmpty();
}


bool MDocument::loadScene(QString filename)
{
 if(!filename.isNull() && !filename.isEmpty())
 {
  if(!MItemReader::get()->read(filename.utf8()))
  {
   QMessageBox::critical(MMainWindow::get(), QString("File Parse Error"), MItemReader::get()->getErrorProtocol(), QMessageBox::Ok, QMessageBox::NoButton);
   
   return false;
  }
 }
 else
 {
  return false;
 }
 
 return true;
}


void MDocument::lock(void)
{
 _lock = true;
}


void MDocument::unlock(void)
{
 _lock = false;
}


MDocument *MDocument::get(void)
{
 if(!_this)
 {
  _this = new MDocument();
 }
 
 return _this;
}


void MDocument::push(QPtrList<QString> &list, QString *value)
{
 list.prepend(value);
}


QString *MDocument::pop(QPtrList<QString> &list)
{
 list.first();
 return list.take();
}
