
/*
 This file is part of moth. moth is a program for creating
 and editing textured 3D models.
 
 Copyright (C) 2004 Peter Uray.

 moth is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 moth is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with moth; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <iostream>

#include <MModelView.h>

MModelView *MModelView::_this = NULL;

MModelView::MModelView(QWidget *parent, const char *name, WFlags f) : QWidget(parent, name, f)
{
 _3dView = new M3DView(this, "3D View");
 _3dView->setBackgroundColor(QColor(160, 160, 160));
 _3dView->hide();
 
 _xyView = new MPlaneView(this, "XY View", MGLView::ZAxis);
 _xyView->setBackgroundColor(QColor(160, 160, 140));
 _xyView->hide();
  
 _yzView = new MPlaneView(this, "YZ View", MGLView::XAxis);
 _yzView->setBackgroundColor(QColor(140, 160, 160));
 _yzView->hide();
 
 _zxView = new MPlaneView(this, "ZX View", MGLView::YAxis);
 _zxView->setBackgroundColor(QColor(160, 140, 160)); 
 _zxView->hide();
 
 _textureView = new MTextureView(this, "Texture View");
 _textureView->setBackgroundColor(QColor(160, 160, 160));
 _textureView->hide();
 
 // view mode is VIEWMODEQUAD initially.
 selectQuadView();
}


MModelView *MModelView::get(void)
{
 return _this;
}


void MModelView::create(QWidget *parent)
{
 _this = new MModelView(parent);
}


void MModelView::repaint(void)
{
 _3dView->updateGL();
 _xyView->updateGL();
 _yzView->updateGL();
 _zxView->updateGL();
 _textureView->updateGL();
}


void MModelView::resizeEvent(QResizeEvent *event)
{
 switch(_viewMode)
 {
  case ViewMode3D: select3DView(); break;
  case ViewModeXY: selectXYView(); break;
  case ViewModeYZ: selectYZView(); break;
  case ViewModeZX: selectZXView(); break;
  case ViewModeQuad: selectQuadView(); break;
  case ViewModeTexture: selectTextureView(); break;
 }
}


void MModelView::select3DView(void)
{
 _viewMode = ViewModeQuad;

 _textureView->hide();

 _3dView->setGeometry(1, 1, width() - 1, 3*height()/4 - 1);
 _3dView->show();
 _3dView->updateGL();
 
 _xyView->setGeometry(1, 3*height()/4 + 1, width() / 3 - 1, height()/4 - 2);
 _xyView->show();
 _xyView->updateGL();

 _yzView->setGeometry(width()/3 + 1, 3*height()/4 + 1, width()/3 - 1, height()/4 - 2);
 _yzView->show();
 _yzView->updateGL();
 
 _zxView->setGeometry(2*width()/3 + 1, 3*height()/4 + 1, width()/3 - 1, height()/4 - 2);
 _zxView->show();
 _zxView->updateGL();
}


void MModelView::selectXYView(void)
{
 _viewMode = ViewModeXY;

 _textureView->hide();
 
 _xyView->setGeometry(1, 1, width() - 1, 3*height()/4 - 1);
 _xyView->show();
 _xyView->updateGL();
 
 _3dView->setGeometry(1, 3*height()/4 + 1, width() / 3 - 1, height()/4 - 2);
 _3dView->show();
 _3dView->updateGL();

 _yzView->setGeometry(width()/3 + 1, 3*height()/4 + 1, width()/3 - 1, height()/4 - 2);
 _yzView->show();
 _yzView->updateGL();
 
 _zxView->setGeometry(2*width()/3 + 1, 3*height()/4 + 1, width()/3 - 1, height()/4 - 2);
 _zxView->show();
 _zxView->updateGL();
}


void MModelView::selectYZView(void)
{
 _viewMode = ViewModeYZ;

 _textureView->hide();
 
 _yzView->setGeometry(1, 1, width() - 1, 3*height()/4 - 1);
 _yzView->show();
 _yzView->updateGL();
 
 _xyView->setGeometry(1, 3*height()/4 + 1, width() / 3 - 1, height()/4 - 2);
 _xyView->show();
 _xyView->updateGL();

 _3dView->setGeometry(width()/3 + 1, 3*height()/4 + 1, width()/3 - 1, height()/4 - 2);
 _3dView->show();
 _3dView->updateGL();
 
 _zxView->setGeometry(2*width()/3 + 1, 3*height()/4 + 1, width()/3 - 1, height()/4 - 2);
 _zxView->show();
 _zxView->updateGL();
}


void MModelView::selectZXView(void)
{
 _viewMode = ViewModeZX;

 _textureView->hide();

 _zxView->setGeometry(1, 1, width() - 1, 3*height()/4 - 1);
 _zxView->show();
 _zxView->updateGL();
 
 _xyView->setGeometry(1, 3*height()/4 + 1, width() / 3 - 1, height()/4 - 2);
 _xyView->show();
 _xyView->updateGL();

 _yzView->setGeometry(width()/3 + 1, 3*height()/4 + 1, width()/3 - 1, height()/4 - 2);
 _yzView->show();
 _yzView->updateGL();
 
 _3dView->setGeometry(2*width()/3 + 1, 3*height()/4 + 1, width()/3 - 1, height()/4 - 2);
 _3dView->show();
 _3dView->updateGL();
}


void MModelView::selectQuadView(void)
{
 _viewMode = ViewModeQuad;

 _textureView->hide();

 _3dView->setGeometry(1, 1, width() / 2 - 1, height() / 2 - 1);
 _3dView->show();
 _3dView->updateGL();
 
 _xyView->setGeometry(width() / 2 + 1, 1, width() / 2 - 1, height() / 2 - 1);
 _xyView->show();
 _xyView->updateGL();

 _yzView->setGeometry(1, height() / 2 + 1, width() / 2 - 1, height() / 2 - 1);
 _yzView->show();
 _yzView->updateGL();
 
 _zxView->setGeometry(width() / 2 + 1, height() / 2 + 1, width() / 2 - 1, height() / 2 - 1);
 _zxView->show();
 _zxView->updateGL();
}


void MModelView::selectTextureView(void)
{
 _viewMode = ViewModeTexture;

 _textureView->hide();
 _3dView->hide();
 _xyView->hide();
 _yzView->hide();
 _zxView->hide();

 _textureView->setGeometry(1, 1, width() - 1, height() - 1);
 _textureView->show();
 _textureView->updateGL();
}


M3DView &MModelView::get3DView(void)
{
 return *_3dView;
}


MPlaneView &MModelView::getXYView(void)
{
 return *_xyView;
}


MPlaneView &MModelView::getYZView(void)
{
 return *_yzView;
}


MPlaneView &MModelView::getZXView(void)
{
 return *_zxView;
}


MTextureView &MModelView::getTextureView(void)
{
 return *_textureView;
}


MModelView::ViewMode MModelView::getViewMode(void)
{
 return _viewMode;
}
