
/*
 This file is part of moth. moth is a program for creating
 and editing textured 3D models.
 
 Copyright (C) 2004 Peter Uray.

 moth is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 moth is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with moth; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <qcursor.h>

#include <moth/MVector.h>

#include <MPlaneView.h>
#include <MStructureView.h>
#include <MMainWindow.h>
#include <MPreferences.h>

#include <moth/MGlobalLight.h>

#include <MItemManipulator.h>
#include <moth/MModelManipulator.h>
#include <moth/MShapeManipulator.h>
#include <moth/MLightSourceManipulator.h>


MPlaneView::MPlaneView(QWidget *parent, const char *name, MAxis viewingAxis) : MGLView(parent, name)
{
 _axis = viewingAxis;
 
 _position = O;
 _scale = 1.0;
 
 _oldMouseX = 0;
 _oldMouseY = 0;
  
 setFocusPolicy(QWidget::StrongFocus);
 setMouseTracking(true);
}


void MPlaneView::renderView(void)
{
 float orthoX1, orthoY1, orthoX2, orthoY2;
 float defaultSize;
 float viewAspect;
 int gx, gy, gz;
 MVector p1, p2, p3;
 
 if(!isVisible()) return;
 
 defaultSize = MCoordinateSystem::get()->getSize();

 viewAspect = ((float)width()) / ((float)height()); 

 if(width() < height())
 {
  orthoX1 = -_scale*defaultSize*viewAspect;
  orthoY1 = -_scale*defaultSize;
  orthoX2 = _scale*defaultSize*viewAspect;
  orthoY2 = _scale*defaultSize;
 }
 else
 {
  orthoX1 = -_scale*defaultSize;
  orthoY1 = -_scale*defaultSize/viewAspect;
  orthoX2 = _scale*defaultSize;
  orthoY2 = _scale*defaultSize/viewAspect;
 }
 
 float nearClippingDistance = MPreferences::get()->getNCD();
 float farClippingDistance = MPreferences::get()->getFCD();
 
 _camera.createOrthoView(orthoX1, orthoY1, orthoX2, orthoY2, nearClippingDistance, farClippingDistance, width(), height());
 
 p1 = "1.0 0.0 0.0"; p1[0] += 5.0*MCoordinateSystem::get()->getSize();
 p2 = "0.0 -1.0 0.0"; p2[1] -= 5.0*MCoordinateSystem::get()->getSize();
 p3 = "0.0 0.0 1.0"; p3[2] += 5.0*MCoordinateSystem::get()->getSize();

 switch(_axis)
 {
  case XAxis:
    _camera.setPosition(O);
    _camera.rotate(90.0, e2);
    _camera.rotate(90.0, e1);
    _camera.setPosition(p1 + _position);
  break;
  
  case YAxis:
    _camera.setPosition(O);
    _camera.rotate(90.0, e1);
    _camera.setPosition(p2 + _position);
  break;
  
  case ZAxis:
    _camera.setPosition(p3 + _position);
  break;
  
  default:
  
  break;
 }
 
 // render the scene ..
 _camera.render();
 
 glEnable(GL_DEPTH_TEST);
 
 glMatrixMode(GL_MODELVIEW); 
 glViewport(0, 0, (GLint)width(), (GLint)height());
 
 glClearColor(_backgroundColor.red() / 256.0, _backgroundColor.green() / 256.0, _backgroundColor.blue() / 256.0, 0.0);
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 
 glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 glEnable(GL_LIGHTING);
 glDepthFunc(GL_LEQUAL);

 MGlobalLight::get()->render();
 
 MSceneGraph::get()->updateData();
 MSceneGraph::get()->render();
  
 gx = MCoordinateSystem::get()->getGridModeXY(); 
 gy = MCoordinateSystem::get()->getGridModeYZ(); 
 gz = MCoordinateSystem::get()->getGridModeZX();

 MCoordinateSystem::get()->readParameters();
 MCoordinateSystem::get()->setGridMode(1, 1, 1); 
 MCoordinateSystem::get()->setArrows(false);
 MCoordinateSystem::get()->render();
 MCoordinateSystem::get()->setArrows(true);
 MCoordinateSystem::get()->setGridMode(gx, gy, gz);
 
 MItemManipulator::get()->render();
}



void MPlaneView::onSelectModel(void)
{
 MSceneGraph::get()->setSelectionHotSpot(_newMouseX, height() - _newMouseY);
 MModelView::get()->repaint();
 
 MDocument::get()->setSelectedItem(MSceneGraph::get()->getSelectedItem());
 MModelView::get()->repaint();
 
 MStructureView::get()->setSelectedItem(MSceneGraph::get()->getSelectedItem());
}


void MPlaneView::onSelectShape(void)
{
 
}


void MPlaneView::onSelectLight(void)
{

}



void MPlaneView::onZoomIn(void)
{
 float width, height;
 float fdx, fdy;
 MVector dx;

 width = _camera.getOrthoX2() - _camera.getOrthoX1();
 height = _camera.getOrthoY2() - _camera.getOrthoY1();

 fdx = ((float)(_newMouseX - this->width()/2))/((float)_camera.getViewportWidth())*width;
 fdy = -((float)(_newMouseY - this->height()/2))/((float)_camera.getViewportHeight())*height;
 
 switch(_axis)
 {
  case MGLView::XAxis: dx[0] = 0.0; dx[1] = fdx; dx[2] = fdy; break;
  case MGLView::YAxis: dx[0] = fdx; dx[1] = 0.0; dx[2] = fdy; break;
  case MGLView::ZAxis: dx[0] = fdx; dx[1] = fdy; dx[2] = 0.0; break;
  default: break;
 }
 
 _position += dx;
 _scale *= 0.95;
  
 MModelView::get()->repaint();
}


void MPlaneView::onZoomOut(void)
{
 float width, height;
 float fdx, fdy;
 MVector dx;

 width = _camera.getOrthoX2() - _camera.getOrthoX1();
 height = _camera.getOrthoY2() - _camera.getOrthoY1();

 fdx = ((float)(_newMouseX - this->width()/2))/((float)_camera.getViewportWidth())*width;
 fdy = -((float)(_newMouseY - this->height()/2))/((float)_camera.getViewportHeight())*height;
 
 switch(_axis)
 {
  case MGLView::XAxis: dx[0] = 0.0; dx[1] = fdx; dx[2] = fdy; break;
  case MGLView::YAxis: dx[0] = fdx; dx[1] = 0.0; dx[2] = fdy; break;
  case MGLView::ZAxis: dx[0] = fdx; dx[1] = fdy; dx[2] = 0.0; break;
  default: break;
 }
 
 _position += dx;
 _scale *= 1.05;
  
 MModelView::get()->repaint();
}



// ----------------------------------------------------------------------------
// Move Model
// ----------------------------------------------------------------------------
void MPlaneView::onMoveModelInit(void)
{
 
}


void MPlaneView::onMoveModelPerform(void)
{
 MVector p1, p1_prime, p2;
 MModelManipulator mm;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
 
 p2 = determinePosition3D();
 
 p1 = MDocument::get()->getSelectedModel()->getPosition();
 p1_prime = p1;
 MCoordinateSystem::get()->snap(p1_prime);
    
 switch(_axis)
 {
  case MGLView::XAxis: p2[0] = p1_prime[0]; break;
  case MGLView::YAxis: p2[1] = p1_prime[1]; break;
  case MGLView::ZAxis: p2[2] = p1_prime[2]; break;
  default: break;
 }
    
 MCoordinateSystem::get()->snap(p2);
 
 mm.set(MDocument::get()->getSelectedModel());
 if(_state & RightButton) mm.setChangeGeometry(false);
 mm.move(p2 - p1);

 MModelView::get()->repaint();
}


void MPlaneView::onMoveModelFinish(void)
{
 MDocument::get()->unlock();
}

// ----------------------------------------------------------------------------
// Move Shape
// ----------------------------------------------------------------------------
void MPlaneView::onMoveShapeInit(void)
{

}


void MPlaneView::onMoveShapePerform(void)
{
 MVector p1, p1_prime, p2;
 MShapeManipulator sm;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
 
 p2 = determinePosition3D();
 
 p1 = MDocument::get()->getSelectedShape()->getPosition();
 p1_prime = p1;
 MCoordinateSystem::get()->snap(p1_prime);
    
 switch(_axis)
 {
  case MGLView::XAxis: p2[0] = p1_prime[0]; break;
  case MGLView::YAxis: p2[1] = p1_prime[1]; break;
  case MGLView::ZAxis: p2[2] = p1_prime[2]; break;
  default: break;
 }
    
 MCoordinateSystem::get()->snap(p2);
 
 sm.set(MDocument::get()->getSelectedShape());
 if(_state & RightButton) sm.setChangeGeometry(false);
 sm.move(p2 - p1);

 MModelView::get()->repaint();
}


void MPlaneView::onMoveShapeFinish(void)
{
 MDocument::get()->unlock();
}


// ----------------------------------------------------------------------------
// Move Light
// ----------------------------------------------------------------------------
void MPlaneView::onMoveLightInit(void)
{
 
}


void MPlaneView::onMoveLightPerform(void)
{
 MVector p1, p1_prime, p2;
 MLightSourceManipulator lm;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
 
 p2 = determinePosition3D();
 
 p1 = MDocument::get()->getSelectedLightSource()->getPosition();
 p1_prime = p1;
 MCoordinateSystem::get()->snap(p1_prime);
    
 switch(_axis)
 {
  case MGLView::XAxis: p2[0] = p1_prime[0]; break;
  case MGLView::YAxis: p2[1] = p1_prime[1]; break;
  case MGLView::ZAxis: p2[2] = p1_prime[2]; break;
  default: break;
 }
    
 MCoordinateSystem::get()->snap(p2);
 
 lm.set(MDocument::get()->getSelectedLightSource());
 lm.move(p2 - p1);

 MModelView::get()->repaint();
}


void MPlaneView::onMoveLightFinish(void)
{
 MDocument::get()->unlock();
}



// ----------------------------------------------------------------------------
// Rotate Model
// ----------------------------------------------------------------------------
void MPlaneView::onRotateModelInit(void)
{
 
}


void MPlaneView::onRotateModelPerform(void)
{
 MVector axis;
 float angle;
 MModelManipulator mm;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
 angle = 0.0;
 switch(_axis)
 {
  case MGLView::XAxis:    
    axis = e1;
    angle = -2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleX());
  break;


  case MGLView::YAxis:
    axis = e2;
    angle = 2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleY());
  break;
  
  
  case MGLView::ZAxis:
    axis = e3;
    angle = -2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleZ());
  break;
  
  
  default:
  
  break;
 }
 
 mm.set(MDocument::get()->getSelectedModel());
 if(_state & RightButton) mm.setChangeGeometry(false);
 mm.rotate(angle, axis);
 
 MModelView::get()->repaint();
}


void MPlaneView::onRotateModelFinish(void)
{
 MDocument::get()->unlock();
}


// ----------------------------------------------------------------------------
// Rotate Shape
// ----------------------------------------------------------------------------
void MPlaneView::onRotateShapeInit(void)
{
 
}


void MPlaneView::onRotateShapePerform(void)
{
 MVector axis;
 float angle;
 MShapeManipulator sm;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
 angle = 0.0;
 switch(_axis)
 {
  case MGLView::XAxis:    
    axis = e1;
    angle = -2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleX());
  break;


  case MGLView::YAxis:
    axis = e2;
    angle = 2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleY());
  break;
  
  
  case MGLView::ZAxis:
    axis = e3;
    angle = -2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleZ());
  break;
  
  
  default:
  
  break;
 }
 
 sm.set(MDocument::get()->getSelectedShape());
 if(_state & RightButton) sm.setChangeGeometry(false);
 sm.rotate(angle, axis);
 
 MModelView::get()->repaint();
}


void MPlaneView::onRotateShapeFinish(void)
{
 MDocument::get()->unlock();
}


// ----------------------------------------------------------------------------
// Rotate Light
// ----------------------------------------------------------------------------
void MPlaneView::onRotateLightInit(void)
{
 
}


void MPlaneView::onRotateLightPerform(void)
{
 MVector axis;
 float angle;
 MLightSourceManipulator lm;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
 angle = 0.0;
 switch(_axis)
 {
  case MGLView::XAxis:    
    axis = e1;
    angle = -2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleX());
  break;


  case MGLView::YAxis:
    axis = e2;
    angle = 2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleY());
  break;
  
  
  case MGLView::ZAxis:
    axis = e3;
    angle = -2.0*_mouseDX;
    if(MCoordinateSystem::get()->getSnapMode()) angle = MCoordinateSystem::get()->snapFunc(angle, MCoordinateSystem::get()->getSnapAngleZ());
  break;
  
  
  default:
  
  break;
 }
 
 lm.set(MDocument::get()->getSelectedLightSource()); 
 lm.rotate(angle, axis);
 
 MModelView::get()->repaint();
}


void MPlaneView::onRotateLightFinish(void)
{
 MDocument::get()->unlock();
}



// ----------------------------------------------------------------------------
// Scale model
// ----------------------------------------------------------------------------
void MPlaneView::onScaleModelInit(void)
{
 
}


void MPlaneView::onScaleModelPerform(void)
{
 float s1, s2, s3;
 MVector axis;
 MModelManipulator mm;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
  
 s1 = 1.0; s2 = 1.0; s3 = 1.0;
 
 switch(_axis)
 {
  case MGLView::XAxis:
    fabs(_mouseDX) > fabs(_mouseDY) ? s2 += _mouseDX*0.02 : s3 -= _mouseDY*0.02;
  break;
  
  
  case MGLView::YAxis:
    fabs(_mouseDX) > fabs(_mouseDY) ? s1 += _mouseDX*0.02 : s3 -= _mouseDY*0.02;
  break;
  
  
  case MGLView::ZAxis:
    fabs(_mouseDX) > fabs(_mouseDY) ? s1 += _mouseDX*0.02 : s2 -= _mouseDY*0.02;
  break;
  
  
  default:
  
  break;
 }
 
 mm.set(MDocument::get()->getSelectedModel());
 mm.scale(s1, e1);
 mm.scale(s2, e2);
 mm.scale(s3, e3); 
 MModelView::get()->repaint();
}


void MPlaneView::onScaleModelFinish(void)
{
 MDocument::get()->unlock();
}


// ----------------------------------------------------------------------------
// Scale shape
// ----------------------------------------------------------------------------
void MPlaneView::onScaleShapeInit(void)
{
 
}


void MPlaneView::onScaleShapePerform(void)
{
 float s1, s2, s3;
 MShapeManipulator sm;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
  
 s1 = 1.0; s2 = 1.0; s3 = 1.0;
 
 switch(_axis)
 {
  case MGLView::XAxis:
    fabs(_mouseDX) > fabs(_mouseDY) ? s2 += _mouseDX*0.02 : s3 -= _mouseDY*0.02;
  break;
  
  
  case MGLView::YAxis:
    fabs(_mouseDX) > fabs(_mouseDY) ? s1 += _mouseDX*0.02 : s3 -= _mouseDY*0.02;
  break;
  
  
  case MGLView::ZAxis:
    fabs(_mouseDX) > fabs(_mouseDY) ? s1 += _mouseDX*0.02 : s2 -= _mouseDY*0.02;
  break;
  
  
  default:
  
  break;
 }
 
 sm.set(MDocument::get()->getSelectedShape());
 sm.scale(s1, e1);
 sm.scale(s2, e2);
 sm.scale(s3, e3); 
 
 MModelView::get()->repaint();
}


void MPlaneView::onScaleShapeFinish(void)
{
 MDocument::get()->unlock();
}


// ----------------------------------------------------------------------------
// Vertex and facet manipulation
// ----------------------------------------------------------------------------
void MPlaneView::onSelectVertexInit(void)
{
 MLine ray;
 
 if(!MDocument::get()->getSelectedShape()) return;
  
 MItemManipulator::get()->setEraseVertexTo(false);
 
 if((_state & RightButton))
 {
  MItemManipulator::get()->setEraseVertexTo(true);
 }
  
 if((_state & ControlButton))
 {
  
 }
 else
 { 
  MItemManipulator::get()->clearVertexStack();
 }
 
 ray = _camera.getViewRay(_newMouseX, _newMouseY);
 
 MItemManipulator::get()->selectVertices(ray);
 
 MModelView::get()->repaint();
}


void MPlaneView::onSelectVertexPerform(void)
{
 MLine ray;
 
 if(!MDocument::get()->getSelectedShape()) return;

 ray = _camera.getViewRay(_newMouseX, _newMouseY);
 
 MItemManipulator::get()->selectVertices(ray);
 
 MModelView::get()->repaint();
}


void MPlaneView::onSelectVertexFinish(void)
{
 MModelView::get()->repaint();
}


void MPlaneView::onCreateVertex(void)
{
 MVertex *vertex;
 MVector position;
 
 if(MDocument::get()->getSelectedShape())
 {
  MDocument::get()->commit();
 
  position = determinePosition3D();

  MCoordinateSystem::get()->snap(position);
  
  vertex = MDocument::get()->getSelectedShape()->newVertex();
  vertex->setPosition(position);
 }
  
 MModelView::get()->repaint();
}


void MPlaneView::onCreateTriangle(void)
{
 int i1, i2, i3;
 MLine ray;
 
 if(!MDocument::get()->getSelectedShape()) return; 

 if(MDocument::get()->getSelectedShape())
 {
  if(MItemManipulator::get()->getVertexStackSize() < 3)
  {
   MItemManipulator::get()->setEraseVertexTo(false);
 
   ray = _camera.getViewRay(_newMouseX, _newMouseY);
 
   MItemManipulator::get()->selectVertices(ray);
  }
  
  if(MItemManipulator::get()->getVertexStackSize() >= 3)
  {
   i1 = MItemManipulator::get()->getVertexStackEntry(0);
   i2 = MItemManipulator::get()->getVertexStackEntry(1);
   i3 = MItemManipulator::get()->getVertexStackEntry(2);

   MDocument::get()->commit();
   MDocument::get()->getSelectedShape()->addFacet(new MFacet(i1, i2, i3));
   MItemManipulator::get()->clearVertexStack();
  }
  
  MModelView::get()->repaint();
 }
}


void MPlaneView::onMoveVertexInit(void)
{
 if(!MDocument::get()->getSelectedShape()) return;

 if(MItemManipulator::get()->getVertexStackSize() == 0)
 {
  onSelectVertexInit();
 }
 if(MItemManipulator::get()->getVertexStackSize() == 0) return;
}


void MPlaneView::onMoveVertexPerform(void)
{
 unsigned int i;
 MVector dx, newPosition;
 MVertex *v;
 float width, height;
 
 float fdx, fdy;
 
 if(!MDocument::get()->getSelectedShape()) return;
 
 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }

 width = _camera.getOrthoX2() - _camera.getOrthoX1();
 height = _camera.getOrthoY2() - _camera.getOrthoY1();

 fdx = ((float)_mouseDX)/((float)_camera.getViewportWidth())*width;
 fdy = -((float)_mouseDY)/((float)_camera.getViewportHeight())*height;

 switch(_axis)
 {
  case MGLView::XAxis: dx[0] = 0.0; dx[1] = fdx; dx[2] = fdy; break;
  case MGLView::YAxis: dx[0] = fdx; dx[1] = 0.0; dx[2] = fdy; break;
  case MGLView::ZAxis: dx[0] = fdx; dx[1] = fdy; dx[2] = 0.0; break;
  default: break;
 }

 for(i = 0; i < MItemManipulator::get()->getVertexStackSize(); i++)
 {
  v = MDocument::get()->getSelectedShape()->getVertex(MItemManipulator::get()->getVertexStackEntry(i));
 
  newPosition = v->getPosition() + dx;
  v->setPosition(newPosition);   
 }
 MModelView::get()->repaint();
}


void MPlaneView::onMoveVertexFinish(void)
{
 unsigned int i;
 MVector newPosition;
 MVertex *v;

 MDocument::get()->unlock();
 for(i = 0; i < MItemManipulator::get()->getVertexStackSize(); i++)
 {
  v = MDocument::get()->getSelectedShape()->getVertex(MItemManipulator::get()->getVertexStackEntry(i));
 
  newPosition = v->getPosition();
  MCoordinateSystem::get()->snap(newPosition);
  v->setPosition(newPosition);   
 }
 MModelView::get()->repaint();
}




void MPlaneView::onRotateVertexInit(void)
{

}


void MPlaneView::onRotateVertexPerform(void)
{
 float angle;
 unsigned int i;
 MVector pos, newPosition;
 MVector axis;
 MVertex *v;
 MMatrix R;

 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
 angle = 3.0*_mouseDX;

 switch(_axis)
 {
  case MGLView::XAxis:
    R.loadRotation(angle, e1);
	break;
  
	case MGLView::YAxis:
	  R.loadRotation(angle, e2);
	break;
  
	case MGLView::ZAxis:
	  R.loadRotation(angle, e3);
	break;
  
	default:
	  angle = 0.0;
	break;
 } 

 pos = MDocument::get()->getSelectedItem()->getPosition();
 for(i = 0; i < MItemManipulator::get()->getVertexStackSize(); i++)
 {
  v = MDocument::get()->getSelectedShape()->getVertex(MItemManipulator::get()->getVertexStackEntry(i));
  
  newPosition = R*(v->getPosition() - pos) + pos;
  v->setPosition(newPosition);   
 } 
 MModelView::get()->repaint();
}


void MPlaneView::onRotateVertexFinish(void)
{
 unsigned int i;
 MVector newPosition;
 MVertex *v;

 MDocument::get()->unlock();
 for(i = 0; i < MItemManipulator::get()->getVertexStackSize(); i++)
 {
  v = MDocument::get()->getSelectedShape()->getVertex(MItemManipulator::get()->getVertexStackEntry(i));
 
  newPosition = v->getPosition();
  MCoordinateSystem::get()->snap(newPosition);
  v->setPosition(newPosition);   
 }
 MModelView::get()->repaint();
}


void MPlaneView::onScaleVertexInit(void)
{

}


void MPlaneView::onScaleVertexPerform(void)
{
 MVector dir;
 float way;
 unsigned int i;
 MVector newPosition;
 MVertex *v;
 MVector pos;
 MMatrix S;

 if(_mouseDS > 0.0)
 {
  MDocument::get()->commit();
  MDocument::get()->lock();
 }
 
 way = 0.1*_mouseDX;

 switch(_axis)
 {
  case MGLView::XAxis:
    S.loadDiagonal(1.0, 1.0 + way, 1.0 + way);
	break;
  
	case MGLView::YAxis:
	  S.loadDiagonal(1.0 + way, 1.0, 1.0 + way);
	break;
  
	case MGLView::ZAxis:
	  S.loadDiagonal(1.0 + way, 1.0 + way, 1.0);
	break;
  
	default:
    S.loadDiagonal(1.0, 1.0, 1.0);
	break;
 }
 
 
 pos = MDocument::get()->getSelectedItem()->getPosition();
 for(i = 0; i < MItemManipulator::get()->getVertexStackSize(); i++)
 {
  v = MDocument::get()->getSelectedShape()->getVertex(MItemManipulator::get()->getVertexStackEntry(i));
  
  newPosition = S*(v->getPosition() - pos) + pos;
  v->setPosition(newPosition);   
 } 
 MModelView::get()->repaint();
}


void MPlaneView::onScaleVertexFinish(void)
{
 unsigned int i;
 MVector newPosition;
 MVertex *v;

 MDocument::get()->unlock();
 for(i = 0; i < MItemManipulator::get()->getVertexStackSize(); i++)
 {
  v = MDocument::get()->getSelectedShape()->getVertex(MItemManipulator::get()->getVertexStackEntry(i));
 
  newPosition = v->getPosition();
  MCoordinateSystem::get()->snap(newPosition);
  v->setPosition(newPosition);   
 }
 MModelView::get()->repaint();
}


void MPlaneView::onMagnetInit(void)
{

}


void MPlaneView::onMagnetPerform(void)
{

}


void MPlaneView::onMagnetFinish(void)
{

}


void MPlaneView::onFlipEdge(void)
{
 int i1, i2, i3, i4;
 int ti1, ti2, ti3, ti4, ti5, ti6;
 
 MLine ray;
 QPtrList<MFacet> flipFacets;
 MFacet *f;
  
 if(!MDocument::get()->getSelectedShape()) return; 

 if(MDocument::get()->getSelectedShape())
 {
  if(MItemManipulator::get()->getVertexStackSize() < 2)
  {
   MItemManipulator::get()->setEraseVertexTo(false);
 
   ray = _camera.getViewRay(_newMouseX, _newMouseY);
 
   MItemManipulator::get()->selectVertices(ray);
  }
  
  if(MItemManipulator::get()->getVertexStackSize() >= 2)
  {
	 i1 = MItemManipulator::get()->getVertexStackEntry(0);
   i2 = MItemManipulator::get()->getVertexStackEntry(1);

   QPtrList<MFacet> &facetList = MDocument::get()->getSelectedShape()->getFacetList();

   for(f = facetList.first(); f; f = facetList.next())
   {
    if(f->containsVertex(i1) && f->containsVertex(i2))
    {
     flipFacets.append(f);
    }
   }
   
   if(flipFacets.count() >= 2)
   {
 	  MDocument::get()->commit();

    i3 = flipFacets.at(0)->remainingVertex(i1, i2);
    i4 = flipFacets.at(1)->remainingVertex(i1, i2);
    
    ti1 = flipFacets.at(0)->vertex2TextureVertex(i1);
    ti2 = flipFacets.at(0)->vertex2TextureVertex(i3);
    ti3 = flipFacets.at(1)->vertex2TextureVertex(i4);
    
    ti4 = flipFacets.at(1)->vertex2TextureVertex(i2);
    ti5 = flipFacets.at(1)->vertex2TextureVertex(i4);
    ti6 = flipFacets.at(0)->vertex2TextureVertex(i3);
    

    flipFacets.at(0)->setVertexRef(0, i1);
    flipFacets.at(0)->setVertexRef(1, i3);
    flipFacets.at(0)->setVertexRef(2, i4);
    
    flipFacets.at(0)->setTextureVertexRef(0, ti1);
    flipFacets.at(0)->setTextureVertexRef(1, ti2);
    flipFacets.at(0)->setTextureVertexRef(2, ti3);
    
    
    flipFacets.at(1)->setVertexRef(0, i2);
    flipFacets.at(1)->setVertexRef(1, i4);
    flipFacets.at(1)->setVertexRef(2, i3);
    
    flipFacets.at(1)->setTextureVertexRef(0, ti4);
    flipFacets.at(1)->setTextureVertexRef(1, ti5);
    flipFacets.at(1)->setTextureVertexRef(2, ti6);
   }
   
   MItemManipulator::get()->clearVertexStack();
  }
  
  MModelView::get()->repaint();
 }
}











void MPlaneView::onSnapToMaster(void)
{
 int i, j;
 MLine ray;
 
 if(!MDocument::get()->getSelectedShape()) return;
 if(MDocument::get()->getSelectedShape() == MDocument::get()->getMaster()) return;

 if(MDocument::get()->getSelectedShape())
 {
  ray = _camera.getViewRay(_newMouseX, _newMouseY);
 
  if(MItemManipulator::get()->getVertexStackSize() < 1)
  {
   MItemManipulator::get()->setEraseVertexTo(false);
   MItemManipulator::get()->selectVertices(ray);
  }
  
  else
  {
   i = MItemManipulator::get()->getVertexStackEntry(0);
   j = MItemManipulator::get()->selectMasterVertex(ray);
   
   if(j >= 0)
   {
	  MDocument::get()->commit();
    MDocument::get()->getSelectedShape()->getVertex(i)->setPosition(MDocument::get()->getMaster()->getVertex(j)->getPosition());
   }
 
   MItemManipulator::get()->clearVertexStack();
  }
  
  MModelView::get()->repaint();
 }
}














// ----------------------------------------------------------------------------
// View transformation
// ----------------------------------------------------------------------------
void MPlaneView::onRotateView(void)
{
 // Not available in plane views ..
}


void MPlaneView::onZoomView(void)
{
 if(_mouseDY > 0)
 {
  _scale *= 0.95;
  MModelView::get()->repaint();
 }
 if(_mouseDY < 0)
 { 
  _scale *= 1.05;
  updateGL();
 }  
}


void MPlaneView::onMoveView(void)
{
 float lambda = MCoordinateSystem::get()->getSize() / 100.0;
    
 switch(_axis)
 {
  case XAxis: _position[1] -= lambda*(float)_mouseDX; _position[2] += lambda*(float)_mouseDY; break;
  case YAxis: _position[0] -= lambda*(float)_mouseDX; _position[2] += lambda*(float)_mouseDY; break;
  case ZAxis: _position[0] -= lambda*(float)_mouseDX; _position[1] += lambda*(float)_mouseDY; break;
  default: break;
 }  
 updateGL();
}


void MPlaneView::onResetView(void)
{
 _position = O;
 _scale = 1.0;
 
 updateGL();
}


QString MPlaneView::mousePosition3D(void)
{
 MVector v;
 v = determinePosition3D();
 return QString("X: %1,  Y: %2,  Z: %3").arg((float)v[0], 0, 'f', 6).arg((float)v[1], 0, 'f', 6).arg((float)v[2], 0, 'f', 6);
}


MVector MPlaneView::determinePosition3D(void)
{
 MLine viewRay;
 MVector position;
 
 viewRay = _camera.getViewRay(_newMouseX, _newMouseY);
 position = viewRay[0];
 
 switch(_axis)
 {
  case XAxis: position[0] = MCoordinateSystem::get()->getGridX(); break;
  case YAxis: position[1] = MCoordinateSystem::get()->getGridY(); break;
  case ZAxis: position[2] = MCoordinateSystem::get()->getGridZ(); break;
  default: break;
 }
 
 MCoordinateSystem::get()->snap(position);
  
 return position;
}
