
/*
 This file is part of moth. moth is a program for creating
 and editing textured 3D models.
 
 Copyright (C) 2004 Peter Uray.

 moth is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 moth is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with moth; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <stdlib.h>

#include <qstring.h>
#include <qfile.h>
#include <qtextstream.h>

#include <MPreferences.h>


/// A global singleton instance.
MPreferences *MPreferences::_this = NULL;


MPreferences::MPreferences(void)
{
 setToDefault();
}


MPreferences::~MPreferences(void)
{

}


void MPreferences::initialize(void)
{
 MPreferences::get()->setDocumentFile("none");
}


MPreferences *MPreferences::get(void)
{
 if(_this == NULL) _this = new MPreferences();
 return _this;
}


bool MPreferences::writePreferences(void)
{
 QFile preferenceFile(getPreferenceFileLocation());
 QTextStream os;
 
 if(!preferenceFile.open(IO_WriteOnly))
 {
  return false;
 }
 
 os.setDevice(&preferenceFile);

 os << "moth v1.0 Configuration File - do not edit\n";
 
 // General tab
 os << "UNDOSTEPS=" << _undoSteps << "\n";
 os << "PROJECTLOCATION=" << _projectLocation << "\n";
   
 // Grid / units tab   
 os << "GRIDPOS=" << _gridX << " " << _gridY << " " << _gridZ << "\n";
 os << "GRIDDISTANCE=" << _gridDX << " " << _gridDY << " " << _gridDZ << "\n";
 os << "GRIDSNAP=" << _snapX << " " << _snapY << " " << _snapZ << "\n";
 os << "ANGLESNAP=" << _snapAngleX << " " << _snapAngleY << " " << _snapAngleZ << "\n";

 os << "DISPLAYXY=" << _displayXY << "\n";
 os << "DISPLAYYZ=" << _displayYZ << "\n";
 os << "DISPLAYZX=" << _displayZX << "\n";
 os << "DISPLAYAXES=" << _displayAxes << "\n";
   
 os << "SNAP=" << _snap << "\n";
 os << "SNAPANGLES=" << _snapAngles << "\n";
 
 os << "GRIDSIZE=" << _size << "\n";
 os << "TEXGRIDSIZE=" << _texSize << "\n";   
   
 // Camera tab
 os << "CAMERAANGLE=" << _angle << "\n";
 os << "NCD="<< _ncd << "\n";
 os << "FCD=" << _fcd << "\n";
   
 os << "DEFAULTRADIUS=" << _rDefault << "\n";
 os << "DEFAULTPHI=" << _phiDefault << "\n";
 os << "DEFAULTTHETA=" << _thetaDefault << "\n";
 
 preferenceFile.close();
 
 return true;
}


bool MPreferences::readPreferences(void)
{
 float x, y, z;
 QString lineBuffer;
 QFile preferenceFile(getPreferenceFileLocation());
 QTextStream os;
 
 setToDefault();
 if(!preferenceFile.open(IO_ReadOnly))
 {
  return false;
 }
 os.setDevice(&preferenceFile);
 
 lineBuffer = os.readLine(); // Read the header line and pipe it to /dev/null ;)
 
 // General tab
 lineBuffer = os.readLine(); _undoSteps = extractParami(lineBuffer.utf8());
 lineBuffer = os.readLine(); _projectLocation = extractParams(lineBuffer.utf8());
  
 // Grid / units tab
 lineBuffer = os.readLine(); extractParam3f(lineBuffer.utf8(), &x, &y, &z);
 _gridX = x; _gridY = y; _gridZ = z;
 
 lineBuffer = os.readLine(); extractParam3f(lineBuffer.utf8(), &x, &y, &z);
 _gridDX = x; _gridDY = y; _gridDZ = z;
 
 lineBuffer = os.readLine(); extractParam3f(lineBuffer.utf8(), &x, &y, &z);
 _snapX = x; _snapY = y; _snapZ = z;
 
 lineBuffer = os.readLine(); extractParam3f(lineBuffer.utf8(), &x, &y, &z);
 _snapAngleX = x; _snapAngleY = y; _snapAngleZ = z;
 
 
 lineBuffer = os.readLine(); _displayXY = (bool)extractParami(lineBuffer.utf8());
 lineBuffer = os.readLine(); _displayYZ = (bool)extractParami(lineBuffer.utf8());
 lineBuffer = os.readLine(); _displayZX = (bool)extractParami(lineBuffer.utf8());
 lineBuffer = os.readLine(); _displayAxes = (bool)extractParami(lineBuffer.utf8());
 lineBuffer = os.readLine(); _snap = (bool)extractParami(lineBuffer.utf8());
 lineBuffer = os.readLine(); _snapAngles = (bool)extractParami(lineBuffer.utf8());
 
 lineBuffer = os.readLine(); _size = extractParamf(lineBuffer.utf8());
 lineBuffer = os.readLine(); _texSize = extractParami(lineBuffer.utf8());
 
 // Camera tab
 lineBuffer = os.readLine(); _angle = extractParamf(lineBuffer.utf8());
 lineBuffer = os.readLine(); _ncd = extractParamf(lineBuffer.utf8());
 lineBuffer = os.readLine(); _fcd = extractParamf(lineBuffer.utf8());
 lineBuffer = os.readLine(); _rDefault = extractParamf(lineBuffer.utf8());
 lineBuffer = os.readLine(); _phiDefault = extractParamf(lineBuffer.utf8());
 lineBuffer = os.readLine(); _thetaDefault = extractParamf(lineBuffer.utf8());

 preferenceFile.close();

 return true;
}


void MPreferences::setToDefault(void)
{
 // General tab
 _undoSteps = 10;
 
 _projectLocation = getenv("HOME");
 _projectLocation += "/moth";
   
 // Grid / units tab   
 _gridX = 0.0;
 _gridY = 0.0;
 _gridZ = 0.0;
   
 _gridDX = 1.0;
 _gridDY = 1.0;
 _gridDZ = 1.0;
   
 _snapX = 1.0;
 _snapY = 1.0;
 _snapZ = 1.0;
   
 _snapAngleX = 5.0;
 _snapAngleY = 5.0;
 _snapAngleZ = 5.0;
   
 _displayXY = true;
 _displayYZ = false;
 _displayZX = false;
 _displayAxes = true;
   
 _snap = false;
 _snapAngles = false;
     
 _size = 10.0;
 _texSize = 5;
   
 // Camera tab
 _angle = 45.0;
 _ncd = 0.1;
 _fcd = 5000.0;
   
 _rDefault = 30.0;
 _phiDefault = 45.0;
 _thetaDefault = 60.0;
}



QString MPreferences::getPreferenceFileLocation(void)
{
 QString preferenceFile;
 
 preferenceFile = getenv("HOME");
 preferenceFile += "/.moth.rc";
 
 return preferenceFile;
}


int MPreferences::extractParami(const char *line)
{
 QString lineStr, valueStr;
 int value;
 
 lineStr = line;
 valueStr = lineStr.section("=", 1);

 sscanf(valueStr.utf8(), "%i", &value);
 
 return value;
}


float MPreferences::extractParamf(const char *line)
{
 QString lineStr, valueStr;
 float value;
 
 lineStr = line;
 valueStr = lineStr.section("=", 1);
 
 sscanf(valueStr.utf8(), "%f", &value);
 
 return value;
}


void MPreferences::extractParam3f(const char *line, float *x, float *y, float *z)
{
 QString lineStr, valueStr;
 
 lineStr = line;
 valueStr = lineStr.section("=", 1);
 
 sscanf(valueStr.utf8(), "%f %f %f", x, y, z);
}


QString MPreferences::extractParams(const char *line)
{ 
 QString lineStr, valueStr;
 
 lineStr = line;
 valueStr = lineStr.section("=", 1);
  
 return valueStr;
}
