
/*
 This file is part of moth. moth is a program for creating
 and editing textured 3D models.
 
 Copyright (C) 2004 Peter Uray.

 moth is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 moth is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with moth; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <string.h>
#include <iostream>

#include <qsplitter.h>
#include <qtabwidget.h>
#include <qheader.h>
#include <qvbox.h>
#include <qapplication.h>
#include <qlistview.h>
#include <qclipboard.h>
#include <qdragobject.h>

#include <moth/MGlobalLight.h>

#include <moth/MItemReader.h>
#include <moth/MPlaneCreator.h>
#include <moth/MSphereCreator.h>
#include <moth/MDiskCreator.h>
#include <moth/MPrismCreator.h>
#include <moth/MTorusCreator.h>
#include <moth/MHemisphereCreator.h>

#include <MElevationCreator.h>
#include <MSweepRotateCreator.h>
#include <MJitterCreator.h>

#include <moth/MModelManipulator.h>
#include <moth/MShapeManipulator.h>
#include <moth/MLightSourceManipulator.h>

#include <MDocument.h>
#include <MMainWindow.h>
#include <MStructureView.h>
#include <MPositionDialog.h>
#include <MItemManipulator.h>
#include <MCreatorDialog.h>
#include <MMaterialDialog.h>
#include <MLightSourceParameterDialog.h>

MStructureView *MStructureView::_this = NULL;

MStructureView::MStructureView(QWidget *parent) : QWidget(parent)
{
 _structureView = new QListView(this);
 _structureView->header()->setClickEnabled(FALSE);
 
 _structureView->addColumn("Structure");
 _structureView->addColumn("Appearance");
 
 _structureView->setRootIsDecorated(TRUE);
 
 // Disable sorting.
 _structureView->setSorting(-1);
 
 _sceneMenu = NULL;
 _modelMenu = NULL;
 _shapeMenu = NULL;
 _lightsourceMenu = NULL;
 _appearanceMenu = NULL;
  
 createContextMenus();
 
 qApp->connect(_structureView, SIGNAL(selectionChanged()), this, SLOT(selectionChanged()));
 qApp->connect(_structureView, SIGNAL(rightButtonPressed(QListViewItem *, const QPoint &, int)), this, SLOT(rightButtonPressed(QListViewItem *, const QPoint &, int)));

 qApp->connect(_structureView, SIGNAL(expanded(QListViewItem *)), this, SLOT(expanded(QListViewItem *)));
 qApp->connect(_structureView, SIGNAL(collapsed(QListViewItem *)), this, SLOT(collapsed(QListViewItem *)));
 
 qApp->connect(_structureView, SIGNAL(itemRenamed(QListViewItem *, int, const QString &)), this, SLOT(itemRenamed(QListViewItem *, int, const QString &)));

 update();
}


void MStructureView::create(QWidget *parent)
{
 _this = new MStructureView(parent);
}


MStructureView *MStructureView::get(void)
{
 return _this;
}


void MStructureView::createContextMenus(void)
{
 _globalLightMenu = new QPopupMenu(this);
 _globalLightMenu->setCheckable(true);
  
 _globalLightMenu->insertItem("off", this, SLOT(onGlobalLightSelect(int)), 0, 0);
 _globalLightMenu->insertSeparator();
 _globalLightMenu->insertItem("dark red", this, SLOT(onGlobalLightSelect(int)), 0, 1);
 _globalLightMenu->insertItem("dark green", this, SLOT(onGlobalLightSelect(int)), 0, 2);
 _globalLightMenu->insertItem("dark blue", this, SLOT(onGlobalLightSelect(int)), 0, 3);
 
 _globalLightMenu->insertSeparator();
 _globalLightMenu->insertItem("red", this, SLOT(onGlobalLightSelect(int)), 0, 4);
 _globalLightMenu->insertItem("green", this, SLOT(onGlobalLightSelect(int)), 0, 5);
 _globalLightMenu->insertItem("blue", this, SLOT(onGlobalLightSelect(int)), 0, 6);
 
 _globalLightMenu->insertSeparator();
 _globalLightMenu->insertItem("light red", this, SLOT(onGlobalLightSelect(int)), 0, 7);
 _globalLightMenu->insertItem("light green", this, SLOT(onGlobalLightSelect(int)), 0, 8);
 _globalLightMenu->insertItem("light blue", this, SLOT(onGlobalLightSelect(int)), 0, 9);
 
 _globalLightMenu->insertSeparator();
 _globalLightMenu->insertItem("dark gray", this, SLOT(onGlobalLightSelect(int)), 0, 10);
 _globalLightMenu->insertItem("gray", this, SLOT(onGlobalLightSelect(int)), 0, 11);
 _globalLightMenu->insertItem("light gray", this, SLOT(onGlobalLightSelect(int)), 0, 12);
 _globalLightMenu->insertItem("white", this, SLOT(onGlobalLightSelect(int)), 0, 13);
 
 _globalLightMenu->setItemChecked(0, true);
 
 
 
 
 _globalBackgroundMenu = new QPopupMenu(this);
 _globalBackgroundMenu->setCheckable(true);
  
 _globalBackgroundMenu->insertItem("default", this, SLOT(onGlobalBackgroundSelect(int)), 0, 0);
 _globalBackgroundMenu->insertSeparator();
 _globalBackgroundMenu->insertItem("dark red", this, SLOT(onGlobalBackgroundSelect(int)), 0, 1);
 _globalBackgroundMenu->insertItem("dark green", this, SLOT(onGlobalBackgroundSelect(int)), 0, 2);
 _globalBackgroundMenu->insertItem("dark blue", this, SLOT(onGlobalBackgroundSelect(int)), 0, 3);
 
 _globalBackgroundMenu->insertSeparator();
 _globalBackgroundMenu->insertItem("red", this, SLOT(onGlobalBackgroundSelect(int)), 0, 4);
 _globalBackgroundMenu->insertItem("green", this, SLOT(onGlobalBackgroundSelect(int)), 0, 5);
 _globalBackgroundMenu->insertItem("blue", this, SLOT(onGlobalBackgroundSelect(int)), 0, 6);
 
 _globalBackgroundMenu->insertSeparator();
 _globalBackgroundMenu->insertItem("light red", this, SLOT(onGlobalBackgroundSelect(int)), 0, 7);
 _globalBackgroundMenu->insertItem("light green", this, SLOT(onGlobalBackgroundSelect(int)), 0, 8);
 _globalBackgroundMenu->insertItem("light blue", this, SLOT(onGlobalBackgroundSelect(int)), 0, 9);
 
 _globalBackgroundMenu->insertSeparator();
 _globalBackgroundMenu->insertItem("black", this, SLOT(onGlobalBackgroundSelect(int)), 0, 10);
 _globalBackgroundMenu->insertItem("gray", this, SLOT(onGlobalBackgroundSelect(int)), 0, 11);
 _globalBackgroundMenu->insertItem("light gray", this, SLOT(onGlobalBackgroundSelect(int)), 0, 12);
 _globalBackgroundMenu->insertItem("white", this, SLOT(onGlobalBackgroundSelect(int)), 0, 13);
 
 _globalBackgroundMenu->setItemChecked(0, true);
 

 // create the scene menu
 if(_sceneMenu != NULL) _sceneMenu->clear();
 else _sceneMenu = new QPopupMenu(this);
 
 // create the document menu.
 _sceneMenu->insertItem("Add Model", this, SLOT(onAddModelToScene()));
 _sceneMenu->insertItem("Add Light Source", this, SLOT(onAddLightSourceToScene()));
 
 _sceneMenu->insertSeparator();
 _sceneMenu->insertItem("Paste", this, SLOT(onPaste()));
 
 _sceneMenu->insertSeparator();
 _sceneMenu->insertItem("Expand", this, SLOT(onExpand()));
 _sceneMenu->insertItem("Collapse", this, SLOT(onCollapse()));
 
 _sceneMenu->insertSeparator();
 _sceneMenu->insertItem("Global illumination", _globalLightMenu);
 _sceneMenu->insertItem("Background", _globalBackgroundMenu);
 
 
 
 // create the model menu
 if(_modelMenu != NULL) _modelMenu->clear();
 else _modelMenu = new QPopupMenu(this);
 
 // fill the model menu.
 
 // create the sub menus first.
 // shape creation menu
 _createShapeMenu = new QPopupMenu(this);
 _createShapeMenu->insertItem("Plane", this, SLOT(onCreatePlane()));
 _createShapeMenu->insertItem("Sphere", this, SLOT(onCreateSphere()));
 _createShapeMenu->insertItem("Disc", this, SLOT(onCreateDisk()));
 _createShapeMenu->insertItem("Prism", this, SLOT(onCreatePrism()));
 _createShapeMenu->insertItem("Torus", this, SLOT(onCreateTorus()));
 _createShapeMenu->insertItem("Hemisphere", this, SLOT(onCreateHemisphere()));
  
 // shape functions.
 _shapeFunctionMenu = new QPopupMenu(this);
 _shapeFunctionMenu->insertItem("Merge Down", this, SLOT(onMergeDown()));
 _shapeFunctionMenu->insertSeparator();
 _shapeFunctionMenu->insertItem("Remove Facets", this, SLOT(onRemoveFacets()));
 _shapeFunctionMenu->insertItem("Remove Degenerate Facets", this, SLOT(onRemoveDegenerateFacets()));
 _shapeFunctionMenu->insertSeparator();
 _shapeFunctionMenu->insertItem("Remove Unused Vertices", this, SLOT(onRemoveUnusedVertices()));
 _shapeFunctionMenu->insertItem("Remove Unused Texture Vertices", this, SLOT(onRemoveUnusedTextureVertices()));
 _shapeFunctionMenu->insertSeparator();
 _shapeFunctionMenu->insertItem("Homogeneous Orientation", this, SLOT(onHomogeneousOrientation()));
 _shapeFunctionMenu->insertItem("Revert Orientation", this, SLOT(onRevertOrientation()));
 _shapeFunctionMenu->insertSeparator();
 _shapeFunctionMenu->insertItem("Texture Projection", this, SLOT(onTextureProjection()));
 _shapeFunctionMenu->insertItem("Clear Texture Mapping", this, SLOT(onClearTextureMapping()));
 _shapeFunctionMenu->insertSeparator();
 _shapeFunctionMenu->insertItem("Clone Selected Vertices", this, SLOT(onCloneSelectedVertices())); 
 _shapeFunctionMenu->insertSeparator();
 _shapeFunctionMenu->insertItem("Snap Vertices", this, SLOT(onSnapVertices()));
 _shapeFunctionMenu->insertItem("Snap Position", this, SLOT(onSnapPosition()));
 _shapeFunctionMenu->insertItem("Align Direction", this, SLOT(onAlignDirection()));
 _shapeFunctionMenu->insertItem("Align Up", this, SLOT(onAlignUp()));
 _shapeFunctionMenu->insertSeparator(); 
 _shapeFunctionMenu->insertItem("Elevate", this, SLOT(onElevateShape()));
 _shapeFunctionMenu->insertItem("Sweep Rotate", this, SLOT(onSweepRotateShape()));
 _shapeFunctionMenu->insertItem("Jitter", this, SLOT(onJitterShape()));
 

 _modelFunctionMenu = new QPopupMenu(this);
 _modelFunctionMenu->insertItem("Snap Position", this, SLOT(onSnapPosition()));
 _modelFunctionMenu->insertItem("Align Direction", this, SLOT(onAlignDirection()));
 _modelFunctionMenu->insertItem("Align Up", this, SLOT(onAlignUp()));
 
 
 _lightFunctionMenu = new QPopupMenu(this);
 _lightFunctionMenu->insertItem("Snap Position", this, SLOT(onSnapPosition()));
 _lightFunctionMenu->insertItem("Align Direction", this, SLOT(onAlignDirection()));
 // _lightFunctionMenu->insertItem("Align Up", this, SLOT(onAlignUp()));
 
 
 _modelMenu->insertItem("Add Model", this, SLOT(onAddModelToModel()));
 _modelMenu->insertItem("Add Shape", this, SLOT(onAddShapeToModel()));
 _modelMenu->insertItem("Add Light Source", this, SLOT(onAddLightSourceToModel())); 
 
 _modelMenu->insertSeparator();
 _modelMenu->insertItem("Cut", this, SLOT(onCut()));
 _modelMenu->insertItem("Copy", this, SLOT(onCopy()));
 _modelMenu->insertItem("Paste", this, SLOT(onPaste()));
 _modelMenu->insertItem("Clone", this, SLOT(onClone()));
 
 _modelMenu->insertSeparator();
 _modelMenu->insertItem("Expand", this, SLOT(onExpand()));
 _modelMenu->insertItem("Collapse", this, SLOT(onCollapse()));
 
 _modelMenu->insertSeparator();
 _modelMenu->insertItem("Create Shape", _createShapeMenu);

 _modelMenu->insertSeparator();
 _modelMenu->insertItem("Set Position", this, SLOT(onSetPosition()));
 _modelMenu->insertItem("Move", this, SLOT(onMove()));
 _modelMenu->insertItem("Rotate", this, SLOT(onRotate()));
 _modelMenu->insertItem("Scale", this, SLOT(onScale()));
 _modelMenu->insertItem("Mirror", this, SLOT(onMirror()));
 
 _modelMenu->insertSeparator();
 _modelMenu->insertItem("Function", _modelFunctionMenu);

 
  
 // create the shape menu.
 if(_shapeMenu != NULL) _shapeMenu->clear();
 else _shapeMenu = new QPopupMenu(this);
 
 // fill the shape menu.
 _shapeMenu->insertItem("Add Light Source", this, SLOT(onAddLightSourceToShape()));
 _shapeMenu->insertSeparator();
 
 _shapeMenu->insertItem("Assign Material", this, SLOT(onAssignMaterial()));
 _shapeMenu->insertSeparator();
 
 _shapeMenu->insertItem("Cut", this, SLOT(onCut()));
 _shapeMenu->insertItem("Copy", this, SLOT(onCopy()));
 _shapeMenu->insertItem("Paste", this, SLOT(onPaste()));
 _shapeMenu->insertItem("Clone", this, SLOT(onClone()));
 _shapeMenu->insertSeparator();
 _shapeMenu->insertItem("Declare Master", this, SLOT(onDeclareMaster()));
 _shapeMenu->insertItem("Clear Master", this, SLOT(onClearMaster()));
 
 _shapeMenu->insertSeparator();
 _shapeMenu->insertItem("Set Position", this, SLOT(onSetPosition()));
 _shapeMenu->insertItem("Move", this, SLOT(onMove()));
 _shapeMenu->insertItem("Rotate", this, SLOT(onRotate()));
 _shapeMenu->insertItem("Scale", this, SLOT(onScale()));
 _shapeMenu->insertItem("Mirror", this, SLOT(onMirror()));
 _shapeMenu->insertSeparator();
 _shapeMenu->insertItem("Function", _shapeFunctionMenu);
 
 
 
 // create the light menu.
 if(_lightsourceMenu != NULL) _lightsourceMenu->clear();
 else _lightsourceMenu = new QPopupMenu(this);
 
 // fill the light menu. 
 _lightsourceMenu->insertItem("Cut", this, SLOT(onCut()));
 _lightsourceMenu->insertItem("Copy", this, SLOT(onCopy()));
 _lightsourceMenu->insertItem("Clone", this, SLOT(onClone()));
 
 _lightsourceMenu->insertSeparator();
 _lightsourceMenu->insertItem("Set Parameters", this, SLOT(onSetLightSourceParameters()));
 
 
 _lightsourceMenu->insertSeparator();
 _lightsourceMenu->insertItem("Set Position", this, SLOT(onSetPosition()));
 _lightsourceMenu->insertItem("Move", this, SLOT(onMove()));
 _lightsourceMenu->insertItem("Rotate", this, SLOT(onRotate()));
  
 _lightsourceMenu->insertSeparator();
 _lightsourceMenu->insertItem("Function", _lightFunctionMenu);
  
  
 // create the appearance menu.
 if(_appearanceMenu != NULL) _appearanceMenu->clear();
 else _appearanceMenu = new QPopupMenu(this);
 
 // fill the appearance menu. 
 _appearanceMenu->insertItem("Textured", this, SLOT(onSetAppearance(int)), 0, (int)MEnum::Textured);
 _appearanceMenu->insertItem("Shaded", this, SLOT(onSetAppearance(int)), 0, (int)MEnum::Shaded);
 _appearanceMenu->insertItem("Wireframe", this, SLOT(onSetAppearance(int)), 0, (int)MEnum::Wireframe);
 _appearanceMenu->insertItem("Invisible", this, SLOT(onSetAppearance(int)), 0, (int)MEnum::Invisible);
}


void MStructureView::onGlobalLightSelect(int index)
{
 int i;

 for(i = 0; i < 14; i++)
 {
  _globalLightMenu->setItemChecked(i, false);  
 }
 
 _globalLightMenu->setItemChecked(index, true);

 switch(index)
 {
  case 0:
    MGlobalLight::get()->set(0.0, 0.0, 0.0);
  break;
  
  case 1:
    MGlobalLight::get()->set(0.3, 0.0, 0.0);
  break;
  
  case 2:
    MGlobalLight::get()->set(0.0, 0.3, 0.0);
  break;
  
  case 3:
    MGlobalLight::get()->set(0.0, 0.0, 0.3);
  break;
  
  
  
  case 4:
    MGlobalLight::get()->set(0.7, 0.0, 0.0);
  break;
  
  case 5:
    MGlobalLight::get()->set(0.0, 0.7, 0.0);
  break;
  
  case 6:
    MGlobalLight::get()->set(0.0, 0.0, 0.7);
  break;
  
  
  
  
  case 7:
    MGlobalLight::get()->set(1.0, 0.0, 0.0);
  break;
  
  case 8:
    MGlobalLight::get()->set(0.0, 1.0, 0.0);
  break;
  
  case 9:
    MGlobalLight::get()->set(0.0, 0.0, 1.0);
  break;
  
  
  case 10:
    MGlobalLight::get()->set(0.3, 0.3, 0.3);
  break;
  
  case 11:
    MGlobalLight::get()->set(0.5, 0.5, 0.5);
  break;
  
  case 12:
    MGlobalLight::get()->set(0.7, 0.7, 0.7);
  break;
  
  case 13:
    MGlobalLight::get()->set(1.0, 1.0, 1.0);
  break;

  default:
  
  break;
 }
 
}



void MStructureView::onGlobalBackgroundSelect(int index)
{
 int i;

 for(i = 0; i < 14; i++)
 {
  _globalBackgroundMenu->setItemChecked(i, false);  
 }
 
 _globalBackgroundMenu->setItemChecked(index, true);

 switch(index)
 {
  case 0:
    MModelView::get()->get3DView().setBackgroundColor(QColor(160, 160, 160));
  break;
  
  case 1:
    MModelView::get()->get3DView().setBackgroundColor(QColor(50, 0, 0));
  break;
  
  case 2:
    MModelView::get()->get3DView().setBackgroundColor(QColor(0, 50, 0));
  break;
  
  case 3:
    MModelView::get()->get3DView().setBackgroundColor(QColor(0, 0, 50));
  break;
  
  
  
  case 4:
    MModelView::get()->get3DView().setBackgroundColor(QColor(150, 0, 0));
  break;
  
  case 5:
    MModelView::get()->get3DView().setBackgroundColor(QColor(0, 150, 0));
  break;
  
  case 6:
    MModelView::get()->get3DView().setBackgroundColor(QColor(0, 0, 150));
  break;
  
    
  
  case 7:
    MModelView::get()->get3DView().setBackgroundColor(QColor(250, 0, 0));
  break;
  
  case 8:
    MModelView::get()->get3DView().setBackgroundColor(QColor(0, 250, 0));
  break;
  
  case 9:
    MModelView::get()->get3DView().setBackgroundColor(QColor(0, 0, 250));
  break;
	
	
    
  case 10:
    MModelView::get()->get3DView().setBackgroundColor(QColor(0, 0, 0));
  break;
  
  case 11:
    MModelView::get()->get3DView().setBackgroundColor(QColor(80, 80, 80));
  break;
  
  case 12:
    MModelView::get()->get3DView().setBackgroundColor(QColor(200, 200, 200));
  break;
  
  case 13:
    MModelView::get()->get3DView().setBackgroundColor(QColor(255, 255, 255));
  break;
    
  
  default:
  
  break;
 }
 
}




void MStructureView::onCut(void)
{
 QListViewItem *titem, *ptitem;
 MSceneGraphItem *item, *pitem;
 
 titem = _structureView->currentItem();
 if(!titem) return;
 item = ((MTreeItem *)titem)->getData();
 if(!item) return;
 
 MDocument::get()->commit();
 
 ptitem = titem->parent();
 pitem = ((MTreeItem *)ptitem)->getData();
 
 /// First copy the item to the clipboard.
 QApplication::clipboard()->setData(new QTextDrag(item->toString()));
 
 // parent item is NULL -> cut item from the scene graph.
 if(pitem == NULL)
 { 
  if(item->getType() == MEnum::Model)
  {
   MSceneGraph::get()->deleteModel((MModel *)item);
  }
  if(item->getType() == MEnum::LightSource)
  {
   MSceneGraph::get()->deleteLightSource((MLightSource *)item);
  }
 }
 else
 {
  if(pitem->getType() == MEnum::Model)
  {
   if(item->getType() == MEnum::Model)
   {
    ((MModel *)pitem)->deleteModel((MModel *)item);
   }
   if(item->getType() == MEnum::LightSource)
   {
    ((MModel *)pitem)->deleteLightSource((MLightSource *)item);
   }
   if(item->getType() == MEnum::Shape)
   {
    ((MModel *)pitem)->deleteShape((MShape *)item);
   }
  }
  
  if(pitem->getType() == MEnum::Shape)
  {
   if(item->getType() == MEnum::LightSource)
   {
    ((MShape *)pitem)->deleteLightSource((MLightSource *)item);
   }
  }
 }
 
 MDocument::get()->setSelectedItem(NULL);
 
 update();
 MModelView::get()->repaint();
}


void MStructureView::onCopy(void)
{
 QListViewItem *titem;
 MSceneGraphItem *item;
 
 titem = _structureView->currentItem();
 if(!titem) return;
 item = ((MTreeItem *)titem)->getData();
 if(!item) return;
 
 QApplication::clipboard()->setData(new QTextDrag(item->toString()));
}


void MStructureView::onPaste(void)
{
 QString data;
 QListViewItem *titem;
 MSceneGraphItem *item;
 
 titem = _structureView->currentItem();
 if(!titem) return;
 item = ((MTreeItem *)titem)->getData();
 
 MDocument::get()->commit();
 
 data = QApplication::clipboard()->text();
 
 // parent item is NULL is handled by the item reader.
 MItemReader::get()->realize(data, item);

 update();
 MModelView::get()->repaint();
}


void MStructureView::onClone(void)
{
 QListViewItem *titem, *ptitem;

 titem = _structureView->currentItem();
 if(!titem) return;  
 ptitem = titem->parent();
 
 onCopy();
 _structureView->setSelected(ptitem, true);
 onPaste();
}



void MStructureView::onCloneSelectedVertices(void)
{
 unsigned int i;
 MVertex *v;

 MShape *shape;
 QListViewItem *titem, *ptitem;
 MSceneGraphItem *item, *pitem;
 
 titem = _structureView->currentItem();
 if(!titem) return;
 item = ((MTreeItem *)titem)->getData();
 if(!item) return;

 ptitem = titem->parent();
 pitem = ((MTreeItem *)ptitem)->getData();

 if(!pitem) return;
 
 if(isShapePtr(item) && isModelPtr(pitem))
 {
  MDocument::get()->commit();
 
  shape = MShapePtr(item)->clone();
  MModelPtr(pitem)->addShape(shape);
  shape->getLightSourceList().clear();
  shape->getVertexList().clear();
  shape->getFacetList().clear();
  shape->getTextureVertexList().clear();
  
  for(i = 0; i < MItemManipulator::get()->getVertexStackSize(); i++)
  {
   v = MDocument::get()->getSelectedShape()->getVertex(MItemManipulator::get()->getVertexStackEntry(i));
   shape->addVertex(new MVertex(*v));
  }
 }
 
 update();
 MModelView::get()->repaint();
}



void MStructureView::onAddModelToScene(void)
{
 MDocument::get()->commit();
 MSceneGraph::get()->newModel();
 update();
 MModelView::get()->repaint();
}


void MStructureView::onAddLightSourceToScene(void)
{
 MDocument::get()->commit();
 MSceneGraph::get()->newLightSource();
 update();
 MModelView::get()->repaint();
}


void MStructureView::onAddModelToModel(void)
{
 if((MDocument::get()->getSelectedItem()) && (MDocument::get()->getSelectedItem()->getType() == MEnum::Model))
 {
  MDocument::get()->commit();
  ((MModel *)(MDocument::get()->getSelectedItem()))->newModel();
  update();
  MModelView::get()->repaint();
 }
}


void MStructureView::onAddShapeToModel(void)
{
 if((MDocument::get()->getSelectedItem()) && (MDocument::get()->getSelectedItem()->getType() == MEnum::Model))
 {
  MDocument::get()->commit();
  ((MModel *)(MDocument::get()->getSelectedItem()))->newShape();
  update();
  MModelView::get()->repaint();
 }
}


void MStructureView::onAddLightSourceToModel(void)
{
 if((MDocument::get()->getSelectedItem()) && (MDocument::get()->getSelectedItem()->getType() == MEnum::Model))
 {
  MDocument::get()->commit();
  ((MModel *)(MDocument::get()->getSelectedItem()))->newLightSource();
  update();
  MModelView::get()->repaint();
 }
}


void MStructureView::onAddLightSourceToShape(void)
{
 if((MDocument::get()->getSelectedItem()) && (MDocument::get()->getSelectedItem()->getType() == MEnum::Shape))
 {
  MDocument::get()->commit();
  ((MShape *)(MDocument::get()->getSelectedItem()))->newLightSource();
  update();
  MModelView::get()->repaint();
 }
}



void MStructureView::onAssignMaterial(void)
{
 MMaterialDialog materialDialog;
 
 MDocument::get()->commit();
 if(materialDialog.exec())
 {
  
 }
}


void MStructureView::onSetAppearance(int mode)
{
 MDocument::get()->commit();
 if((MDocument::get()->getSelectedItem()) && (MDocument::get()->getSelectedItem()->getType() == MEnum::Shape))
 {
  ((MShape *)(MDocument::get()->getSelectedItem()))->setRenderMode((MEnum::RenderMode)mode);
  update();
  MModelView::get()->repaint();
 }
 
 if((MDocument::get()->getSelectedItem()) && (MDocument::get()->getSelectedItem()->getType() == MEnum::Model))
 {
  ((MModel *)(MDocument::get()->getSelectedItem()))->setRenderMode((MEnum::RenderMode)mode);
  update();
  MModelView::get()->repaint();
 }
}



void MStructureView::onSetPosition(void)
{
 MPositionDialog positionDlg("Position: ");
 MVector dx;
 MVector newPosition;
 
 MModelManipulator mm;
 MShapeManipulator sm;
 MLightSourceManipulator lm;
 

 if(MDocument::get()->getSelectedItem())
 {
  positionDlg.setPosition(MDocument::get()->getSelectedItem()->getPosition());
 }
 
 if(positionDlg.exec())
 {
  MDocument::get()->commit();
 
  newPosition = positionDlg.getPosition();
 
  dx = newPosition - MDocument::get()->getSelectedItem()->getPosition();
 
 
  switch(MDocument::get()->getSelectedItem()->getType())
  {
   case MEnum::Model:
  
    mm.set(MDocument::get()->getSelectedItem());
    mm.move(dx);
  
   break;
  
   case MEnum::Shape:

     sm.set(MDocument::get()->getSelectedItem());
     sm.move(dx);
     
   break;
  
   case MEnum::LightSource:
   
     lm.set(MDocument::get()->getSelectedItem());
     lm.move(dx);
  
   break;
  
  
   default:
  
   break;
  }
 }
}


void MStructureView::onMove(void)
{
 MPositionDialog positionDlg("Displacement: ");
 MVector dx;
 MVector newPosition;
 
 MModelManipulator mm;
 MShapeManipulator sm;
 MLightSourceManipulator lm;
 

 if(MDocument::get()->getSelectedItem())
 {
  positionDlg.setPosition(MVector(0.0, 0.0, 0.0));
 }
 
 if(positionDlg.exec())
 {
  MDocument::get()->commit();
 
  dx = positionDlg.getPosition();
 
  switch(MDocument::get()->getSelectedItem()->getType())
  {
   case MEnum::Model:
  
    mm.set(MDocument::get()->getSelectedItem());
    mm.move(dx);
  
   break;
  
   case MEnum::Shape:

     sm.set(MDocument::get()->getSelectedItem());
     sm.move(dx);
     
   break;
  
   case MEnum::LightSource:
   
     lm.set(MDocument::get()->getSelectedItem());
     lm.move(dx);
  
   break;
  
  
   default:
  
   break;
  }
 }
}


void MStructureView::onRotate(void)
{
 MPositionDialog positionDlg("Angles: ");
 MVector angle;
 
 MModelManipulator mm;
 MShapeManipulator sm;
 MLightSourceManipulator lm;
 

 if(MDocument::get()->getSelectedItem())
 {
  positionDlg.setPosition(MVector(0.0, 0.0, 0.0));
 }
 
 if(positionDlg.exec())
 {
  MDocument::get()->commit();
 
  angle = positionDlg.getPosition();
 
  switch(MDocument::get()->getSelectedItem()->getType())
  {
   case MEnum::Model:
  
    mm.set(MDocument::get()->getSelectedItem());
    mm.rotate(angle[2], e3);
    mm.rotate(angle[1], e2);
    mm.rotate(angle[0], e1);
  
   break;
  
   case MEnum::Shape:

     sm.set(MDocument::get()->getSelectedItem());
     sm.rotate(angle[2], e3);
     sm.rotate(angle[1], e2);
     sm.rotate(angle[0], e1);
     
   break;
  
   case MEnum::LightSource:
   
     lm.set(MDocument::get()->getSelectedItem());
     lm.rotate(angle[2], e3);
     lm.rotate(angle[1], e2);
     lm.rotate(angle[0], e1);
  
   break;
  
  
   default:
  
   break;
  }
 }
}


void MStructureView::onScale(void)
{
 MPositionDialog positionDlg("Scaling Factors: ");
 MVector factor;
 
 MModelManipulator mm;
 MShapeManipulator sm;

 if(MDocument::get()->getSelectedItem())
 {
  positionDlg.setPosition(MVector(1.0, 1.0, 1.0));
 }
 
 if(positionDlg.exec())
 {
  MDocument::get()->commit();
 
  factor = positionDlg.getPosition();
 
  switch(MDocument::get()->getSelectedItem()->getType())
  {
   case MEnum::Model:
  
    mm.set(MDocument::get()->getSelectedItem());
    mm.scale(factor[2], e3);
    mm.scale(factor[1], e2);
    mm.scale(factor[0], e1);
  
   break;
  
   case MEnum::Shape:

     sm.set(MDocument::get()->getSelectedItem());
     sm.scale(factor[2], e3);
     sm.scale(factor[1], e2);
     sm.scale(factor[0], e1);
     
   break;
    
   default:
  
   break;
  }
 }
}



void MStructureView::onMirror(void)
{
 MPositionDialog positionDlg("Normal Axis: ");
 MVector axis;
 
 MModelManipulator mm;
 MShapeManipulator sm;

 if(MDocument::get()->getSelectedItem())
 {
  positionDlg.setPosition(MVector(0.0, 0.0, 1.0));
 }
 
 if(positionDlg.exec())
 {
  MDocument::get()->commit();
 
  axis = positionDlg.getPosition();
 
  switch(MDocument::get()->getSelectedItem()->getType())
  {
   case MEnum::Model:
  
    mm.set(MDocument::get()->getSelectedItem());
    mm.mirror(axis, MDocument::get()->getSelectedItem()->getPosition());
  
   break;
  
   case MEnum::Shape:

     sm.set(MDocument::get()->getSelectedItem());
     sm.mirror(axis, MDocument::get()->getSelectedItem()->getPosition());
     
   break;
    
   default:
  
   break;
  }
 }
}



void MStructureView::onSetLightSourceParameters(void)
{
 if(!MDocument::get()->getSelectedLightSource()) return;

 MLightSourceParameterDialog lpd(MDocument::get()->getSelectedLightSource());
 
 MDocument::get()->commit();
 if(lpd.exec())
 {

 }
}



void MStructureView::onCreatePlane(void)
{
 MShape *shape;
 MPlaneCreator pc;
 
 pc.setParameter("X-Tiles", 6);
 pc.setParameter("Y-Tiles", 4);
  
 pc.setParameter("X-Size", 6.0);
 pc.setParameter("Y-Size", 4.0);
  
 pc.setParameter("U-Size", 6.0);
 pc.setParameter("V-Size", 4.0);
 
 MCreatorDialog cd(&pc);
 
 if((MDocument::get()->getSelectedModel()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = pc.create(((MModel *)(MDocument::get()->getSelectedItem())));
   shape->setRenderMode(MEnum::Shaded);

   update();
   MModelView::get()->repaint();
  }
 }
}


void MStructureView::onCreateSphere(void)
{
 MShape *shape;
 MSphereCreator sc;
 
 sc.setParameter("Phi-Tiles", 12);
 sc.setParameter("Theta-Tiles", 6);
  
 sc.setParameter("Radius", 1.0);
  
 sc.setParameter("U-Size", 1.0);
 sc.setParameter("V-Size", 1.0);
 
 MCreatorDialog cd(&sc);
 
 if((MDocument::get()->getSelectedModel()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = sc.create(((MModel *)(MDocument::get()->getSelectedItem())));
   shape->setRenderMode(MEnum::Shaded);

   update();
   MModelView::get()->repaint();
  }
 }
}


void MStructureView::onCreateDisk(void)
{
 MShape *shape;
 MDiskCreator dc;
 
 dc.setParameter("Phi-Tiles", 12);
 dc.setParameter("Theta-Tiles", 6);
 dc.setParameter("Radius", 1.0);  
 dc.setParameter("Texture Radius", 1.0);
 
 MCreatorDialog cd(&dc);
 
 if((MDocument::get()->getSelectedModel()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = dc.create(((MModel *)(MDocument::get()->getSelectedItem())));
   shape->setRenderMode(MEnum::Shaded);

   update();
   MModelView::get()->repaint();
  }
 }
}


void MStructureView::onCreatePrism(void)
{
 MShape *shape;
 MPrismCreator pc;
 
 pc.setParameter("Sectors", 12);
 pc.setParameter("Rings", 6);
 pc.setParameter("Radius1", 1.0);
 pc.setParameter("Radius2", 1.0);
 pc.setParameter("Height", 1.0); 
 pc.setParameter("U-Size", 1.0);
 pc.setParameter("V-Size", 1.0);
 
 MCreatorDialog cd(&pc);
 
 if((MDocument::get()->getSelectedModel()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = pc.create(((MModel *)(MDocument::get()->getSelectedItem())));
   shape->setRenderMode(MEnum::Shaded);

   update();
   MModelView::get()->repaint();
  }
 }
}


void MStructureView::onCreateTorus(void)
{
 MShape *shape;
 MTorusCreator tc;

 tc.setParameter("Segments", 12);
 tc.setParameter("Rings", 6);
 tc.setParameter("Radius", 4.0);
 tc.setParameter("Diameter", 1.0);
 tc.setParameter("U-Size", 1.0);
 tc.setParameter("V-Size", 1.0);

 MCreatorDialog cd(&tc);

 if((MDocument::get()->getSelectedModel()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = tc.create(MDocument::get()->getSelectedModel());
   shape->setRenderMode(MEnum::Shaded);

   update();
   MModelView::get()->repaint();
  }
 }
}


void MStructureView::onCreateHemisphere(void)
{
 MShape *shape;
 MHemisphereCreator hc;
 
 hc.setParameter("Phi-Tiles", 12);
 hc.setParameter("Theta-Tiles", 6);
 hc.setParameter("Radius", 1.0);  
 hc.setParameter("Texture Radius", 1.0);
 
 MCreatorDialog cd(&hc);
 
 if((MDocument::get()->getSelectedModel()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = hc.create(((MModel *)(MDocument::get()->getSelectedItem())));
   shape->setRenderMode(MEnum::Shaded);

   update();
   MModelView::get()->repaint();
  }
 }
}



void MStructureView::onElevateShape(void)
{
 MShape *shape;
 MElevationCreator ec;
 
 ec.setParameter("Rings", 2);
 ec.setParameter("Height", 1.0); 
 ec.setParameter("U-Size", 1.0);
 ec.setParameter("V-Size", 1.0);
 
 MCreatorDialog cd(&ec);
 
 if((MDocument::get()->getSelectedShape()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = ec.create(NULL);
   shape->setRenderMode(MEnum::Shaded);

   update();
   MModelView::get()->repaint();
  }
 }
}


void MStructureView::onSweepRotateShape(void)
{
 MShape *shape;
 MSweepRotateCreator sc;
 
 sc.setParameter("Sectors", 3);
 sc.setParameter("U-Size", 1.0);
 sc.setParameter("V-Size", 1.0);
 
 MCreatorDialog cd(&sc);
 
 if((MDocument::get()->getSelectedShape()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = sc.create(NULL);
   shape->setRenderMode(MEnum::Shaded);

   update();
   MModelView::get()->repaint();
  }
 }
}


void MStructureView::onJitterShape(void)
{
 MShape *shape;
 MJitterCreator jc;
 
 jc.setParameter("Strength", 0.1);
 
 MCreatorDialog cd(&jc);
 
 if((MDocument::get()->getSelectedShape()))
 {
  if(cd.exec())
  {
   MDocument::get()->commit();
  
   shape = jc.create(NULL);
   
   update();
   MModelView::get()->repaint();
  }
 }
}


void MStructureView::onDeclareMaster(void)
{
 MDocument::get()->setMaster();
 MModelView::get()->repaint();
}


void MStructureView::onClearMaster(void)
{
 MDocument::get()->clearMaster();
 MModelView::get()->repaint();
}


void MStructureView::onSnapToMaster(void)
{ 
 std::cout << "snap to master\n";
}


void MStructureView::onMergeDown(void)
{
 MTreeItem *currentItem, *parentItem;
 MShape *s1, *s2;
 MVertex *v;
 MVector *tv;
 MFacet *f;
 int pos;
 int dv, dtv;
  
 if(!MDocument::get()->getSelectedShape()) return;
 
 currentItem = (MTreeItem *)(_structureView->currentItem());
 if(!currentItem) return;
 
 parentItem = (MTreeItem *)(currentItem->parent());
 
 QPtrList<MShape> &shapes = MModelPtr(parentItem->getData())->getShapeList(); 

 s1 = MShapePtr(currentItem->getData());
 pos = shapes.find(s1);
 if(pos < 1) return;

 s2 = shapes.at(pos-1);
 if(s2 == NULL) return;
 
 MDocument::get()->commit();
 
 dv = s2->numVertices();
 dtv = s2->numTextureVertices();
  
 QPtrList<MVertex> &vertices = s1->getVertexList();
 for(v = vertices.first(); v; v = vertices.next())
 {
  s2->addVertex(new MVertex(*v));
 }
 
 QPtrList<MVector> &texture_vertices = s1->getTextureVertexList();
 for(tv = texture_vertices.first(); tv; tv = texture_vertices.next())
 {
  s2->addTextureVertex(new MVector(*tv));
 }
 
 
 
 QPtrList<MFacet> &facets = s1->getFacetList();
 for(f = facets.first(); f; f = facets.next())
 {
  f->shiftReferences(dv, dtv);
  s2->addFacet(new MFacet(*f));
 }
 
 MModelPtr(parentItem->getData())->deleteShape(s1);
 
 MStructureView::get()->update();
 MModelView::get()->repaint();
}





void MStructureView::onRemoveFacets(void)
{
 MFacet *f;
 QPtrList<MFacet> facetsToKill;
 
 if(!MDocument::get()->getSelectedShape()) return;
 MDocument::get()->commit();

 QPtrList<MFacet> &facets = MDocument::get()->getSelectedShape()->getFacetList();
 
 for(f = facets.first(); f; f = facets.next())
 {
  if((MItemManipulator::get()->hasVertexStackEntry(f->getVertexRef(0))) &&
     (MItemManipulator::get()->hasVertexStackEntry(f->getVertexRef(1))) &&
     (MItemManipulator::get()->hasVertexStackEntry(f->getVertexRef(2))))
  {
   facetsToKill.append(f);
  }
 }
 
 for(f = facetsToKill.first(); f; f = facetsToKill.next())
 {
  facets.remove(f);
 }
 
 MItemManipulator::get()->clearVertexStack();
 MModelView::get()->repaint();
}



void MStructureView::onRemoveDegenerateFacets(void)
{
 MFacet *f;
 QPtrList<MFacet> facetsToKill;
 
 if(!MDocument::get()->getSelectedShape()) return;
 MDocument::get()->commit();

 QPtrList<MFacet> &facets = MDocument::get()->getSelectedShape()->getFacetList();
 
 for(f = facets.first(); f; f = facets.next())
 {
  if((f->getVertexRef(0) == f->getVertexRef(1)) ||
     (f->getVertexRef(1) == f->getVertexRef(2)) ||
     (f->getVertexRef(2) == f->getVertexRef(0)))
  
  {
   facetsToKill.append(f);
  }
 }
 
 for(f = facetsToKill.first(); f; f = facetsToKill.next())
 {
  facets.remove(f);
 }
 
 MItemManipulator::get()->clearVertexStack();
 MModelView::get()->repaint();
}


void MStructureView::onRemoveUnusedVertices(void)
{
 MFacet *f;
 int i;
  
 if(!MDocument::get()->getSelectedShape()) return;

 MDocument::get()->commit();
 
 QPtrList<MVertex> &vertices = MDocument::get()->getSelectedShape()->getVertexList();
 QPtrList<MFacet> &facets = MDocument::get()->getSelectedShape()->getFacetList();
 
 bool danglingVertex[vertices.count()];
 
 for(i = 0; i < (int)vertices.count(); i++)
 {
  danglingVertex[i] = true;
 }
 for(f = facets.first(); f; f = facets.next())
 {
  danglingVertex[f->getVertexRef(0)] = false;
  danglingVertex[f->getVertexRef(1)] = false;
  danglingVertex[f->getVertexRef(2)] = false;
 }

 for(i = vertices.count() - 1; i >= 0; i--)
 {
  if(i < 0) break;
  if(danglingVertex[i] == true)
  {
   for(f = facets.first(); f; f = facets.next())
   {
    f->shiftReferences(-1, 0, i+1, 0);
   }
   vertices.remove(i);
  }
 }

 MModelView::get()->repaint();
}


void MStructureView::onRemoveUnusedTextureVertices(void)
{
 MFacet *f;
 int i;
  
 if(!MDocument::get()->getSelectedShape()) return;

 MDocument::get()->commit();
 
 QPtrList<MVector> &vertices = MDocument::get()->getSelectedShape()->getTextureVertexList();
 QPtrList<MFacet> &facets = MDocument::get()->getSelectedShape()->getFacetList();
 
 bool danglingVertex[vertices.count()];
 
 for(i = 0; i < (int)vertices.count(); i++)
 {
  danglingVertex[i] = true;
 }
 for(f = facets.first(); f; f = facets.next())
 {
  danglingVertex[f->getTextureVertexRef(0)] = false;
  danglingVertex[f->getTextureVertexRef(1)] = false;
  danglingVertex[f->getTextureVertexRef(2)] = false;
 }
 
 for(i = vertices.count() - 1; i >= 0; i--)
 {
  if(i < 0) break;
  if(danglingVertex[i] == true)
  {
   for(f = facets.first(); f; f = facets.next())
   {
    f->shiftReferences(0, -1, 0, i);
   }
   vertices.remove(i);
  }
 }
 MModelView::get()->repaint();
}



void MStructureView::onHomogeneousOrientation(void)
{
 if(!MDocument::get()->getSelectedShape()) return;

 MDocument::get()->commit();
 MDocument::get()->getSelectedShape()->homogenizeOrientation();
 
 MModelView::get()->repaint();
}


void MStructureView::onRevertOrientation(void)
{
 if(!MDocument::get()->getSelectedShape()) return;
 
 MDocument::get()->commit();
 MDocument::get()->getSelectedShape()->revertOrientation();
 
 MModelView::get()->repaint();
}


void MStructureView::onTextureProjection(void)
{
 MVector pos;
 MVertex *v;
 MFacet *f;
  
 if(!MDocument::get()->getSelectedShape()) return;

 MDocument::get()->commit();
 
 QPtrList<MVertex> &vertices = MDocument::get()->getSelectedShape()->getVertexList();
 QPtrList<MFacet> &facets = MDocument::get()->getSelectedShape()->getFacetList();
 QPtrList<MVector> &texvertices = MDocument::get()->getSelectedShape()->getTextureVertexList();
 
 for(v = vertices.first(); v; v = vertices.next())
 {
  pos = v->getPosition();
  pos[2] = 0;
  texvertices.append(new MVector(pos));
 }
 
 for(f = facets.first(); f; f = facets.next())
 {
  f->setTextureVertexRef(0, f->getVertexRef(0));
  f->setTextureVertexRef(1, f->getVertexRef(1));
  f->setTextureVertexRef(2, f->getVertexRef(2));
 }
}


void MStructureView::onClearTextureMapping(void)
{
 MVector pos;
 MFacet *f;
  
 if(!MDocument::get()->getSelectedShape()) return;
 MDocument::get()->commit();
 
 QPtrList<MFacet> &facets = MDocument::get()->getSelectedShape()->getFacetList();
 QPtrList<MVector> &texvertices = MDocument::get()->getSelectedShape()->getTextureVertexList();
 
 texvertices.clear();
 
 for(f = facets.first(); f; f = facets.next())
 {
  f->setTextureVertexRef(0, -1);
  f->setTextureVertexRef(1, -1);
  f->setTextureVertexRef(2, -1);
 }
}


void MStructureView::onSnapVertices(void)
{
 bool snap;
 MVertex *v;
 MVector pos;

 if(!MDocument::get()->getSelectedShape()) return;
 MDocument::get()->commit();
 
 QPtrList<MVertex> &vertices = MDocument::get()->getSelectedShape()->getVertexList();
  
 snap = MCoordinateSystem::get()->getSnapMode();
 MCoordinateSystem::get()->setSnap(true);
 for(v = vertices.first(); v; v = vertices.next())
 {
  pos = v->getPosition();
  MCoordinateSystem::get()->snap(pos);
  v->setPosition(pos);
 }
 MCoordinateSystem::get()->setSnap(snap);
 MModelView::get()->repaint();
}


void MStructureView::onSnapPosition(void)
{
 bool snap;
 MVector pos, pos2;
 MSceneGraphItem *item;
 MModelManipulator mm;
 MShapeManipulator sm;
 MLightSourceManipulator lm;
 
 item = MDocument::get()->getSelectedItem();
 if(!item) return;
 MDocument::get()->commit();
 
 pos = item->getPosition();
 pos2 = pos;
 snap = MCoordinateSystem::get()->getSnapMode();
 MCoordinateSystem::get()->setSnap(true);
 MCoordinateSystem::get()->snap(pos);
 MCoordinateSystem::get()->setSnap(snap);
 
 if(isModelPtr(item))
 {
  mm.set(MModelPtr(item));
  mm.move(pos - pos2);
 }
 
 if(isShapePtr(item))
 {
  sm.set(MShapePtr(item));
  sm.move(pos - pos2);
 }
 
 if(isLightSourcePtr(item))
 {
  lm.set(MLightSourcePtr(item));
  lm.move(pos - pos2);
 }
 
 MModelView::get()->repaint();
}


void MStructureView::onAlignDirection(void)
{
 MSceneGraphItem *item;
 MModelManipulator mm;
 MShapeManipulator sm;
 MLightSourceManipulator lm;
 
 item = MDocument::get()->getSelectedItem();
 if(!item) return;
 MDocument::get()->commit();
 
 if(isModelPtr(item))
 {
  mm.set(MModelPtr(item));
  mm.alignDirection(e1);
 }
 
 if(isShapePtr(item))
 {
  sm.set(MShapePtr(item));
  sm.alignDirection(e1);
 }
 
 if(isLightSourcePtr(item))
 {
  lm.set(MLightSourcePtr(item));
  lm.alignDirection(e1);
 }
 
 MModelView::get()->repaint();
}


void MStructureView::onAlignUp(void)
{
 MSceneGraphItem *item;
 MModelManipulator mm;
 MShapeManipulator sm;
 MLightSourceManipulator lm;
 
 item = MDocument::get()->getSelectedItem();
 if(!item) return;
 MDocument::get()->commit();
 
 if(isModelPtr(item))
 {
  mm.set(MModelPtr(item));
  mm.alignUp(e3);
 }
 
 if(isShapePtr(item))
 {
  sm.set(MShapePtr(item));
  sm.alignUp(e3);
 }
 
 if(isLightSourcePtr(item))
 {
  lm.set(MLightSourcePtr(item));
  lm.alignUp(e3);
 }
 
 MModelView::get()->repaint();
}



void MStructureView::setSelectedItem(MSceneGraphItem *item)
{
 if(item == NULL)
 {
  _structureView->clearSelection();
 }
  
 QListViewItemIterator iterator(_structureView);
  
 while(iterator.current())
 {
  if(((MTreeItem *)iterator.current())->getData() == item)
  {
   _structureView->setSelected(iterator.current(), true);
	 _structureView->ensureItemVisible(iterator.current());
  }
  ++iterator;
 }
}



void MStructureView::update(void)
{
 MModel *m;
 QPtrList<MModel> _models;
 MLightSource *l;
 QPtrList<MLightSource> _lights;
 MTreeItem *item, *rootItem;
  
 _structureView->clear();
 
 // first create the scene item.
 rootItem = new MTreeItem(_structureView, "Scene", MEnum::Item);
 rootItem->setData(NULL);
 rootItem->setOpen(true);
 
 _lights = MSceneGraph::get()->getLightSourceList();
 for(l = _lights.first(); l; l = _lights.next())
 {
  item = new MTreeItem(rootItem, l->getName(), MEnum::LightSource);
  item->setData(l);
  item->setRenameEnabled(0, true);
 }
 
 _models = MSceneGraph::get()->getModelList();
 for(m = _models.first(); m; m = _models.next())
 {
  item = new MTreeItem(rootItem, m->getName(), MEnum::Model);
  item->setData(m);
  item->setRenameEnabled(0, true);
  if(m->isExpanded()) item->setOpen(true);
  
  switch(m->getRenderMode())
  {
   case MEnum::Textured: item->setText(1, "textured"); break;
   case MEnum::Shaded: item->setText(1, "shaded"); break;
   case MEnum::Wireframe: item->setText(1, "wireframe"); break;
   case MEnum::Invisible: item->setText(1, "invisible"); break;
  }

  update(m, item);
 }
}


void MStructureView::update(MModel *model, MTreeItem *parent)
{
 MLightSource *l;
 QPtrList<MLightSource> lights;
 MModel *m;
 QPtrList<MModel> models;
 MShape *s;
 QPtrList<MShape> shapes;
 MTreeItem *item, *item2;


 lights = model->getLightSourceList();
 for(l = lights.first(); l; l = lights.next())
 {
  item = new MTreeItem(parent, l->getName(), MEnum::LightSource);
  item->setData(l);
  item->setRenameEnabled(0, true);
 }
  
 models = model->getModelList();
 for(m = models.first(); m; m = models.next())
 {
  item = new MTreeItem(parent, m->getName(), MEnum::Model);
  item->setData(m);
  item->setRenameEnabled(0, true);
  if(m->isExpanded()) item->setOpen(true);
  
  switch(m->getRenderMode())
  {
   case MEnum::Textured: item->setText(1, "textured"); break;
   case MEnum::Shaded: item->setText(1, "shaded"); break;
   case MEnum::Wireframe: item->setText(1, "wireframe"); break;
   case MEnum::Invisible: item->setText(1, "invisible"); break;
  }
  
  update(m, item);
 }
   
 shapes = model->getShapeList();
 for(s = shapes.first(); s; s = shapes.next())
 {
  item = new MTreeItem(parent, s->getName(), MEnum::Shape);
  item->setData(s);
  item->setRenameEnabled(0, true);
  if(s->isExpanded()) item->setOpen(true);
  
  switch(s->getRenderMode())
  {
   case MEnum::Textured: item->setText(1, "textured"); break;
   case MEnum::Shaded: item->setText(1, "shaded"); break;
   case MEnum::Wireframe: item->setText(1, "wireframe"); break;
   case MEnum::Invisible: item->setText(1, "invisible"); break;
  }
  
  lights = s->getLightSourceList();
  for(l = lights.first(); l; l = lights.next())
  {
   item2 = new MTreeItem(item, l->getName(), MEnum::LightSource);
   item2->setData(l);
   item2->setRenameEnabled(0, true);  
  }
 }
}


void MStructureView::rightButtonPressed(QListViewItem *item, const QPoint &point, int column)
{
 if(!item) return;
 if(column < 0) return;
 
 if(column == 0)
 {
  if(((MTreeItem *)item)->getData() == NULL)
  {
   _sceneMenu->exec(point);
   return;
  }
 
  switch(((MTreeItem *)item)->getData()->getType())
  {
   case MEnum::LightSource:
     _lightsourceMenu->exec(point);
   break;
   
   case MEnum::Model:
     _modelMenu->exec(point);
   break;
  
   case MEnum::Shape:
     _shapeMenu->exec(point);
   break;
  
   default:
  
   break;
  }
 }
 
 
 if(column == 1)
 {
  if(((MTreeItem *)item)->getData() != NULL)
  {
   switch(((MTreeItem *)item)->getData()->getType())
   {
    case MEnum::Model:
    case MEnum::Shape:
      _appearanceMenu->exec(point);
    break;
   
    default:
   
    break;
   }
  }
 }
}



void MStructureView::selectionChanged(void)
{
 MTreeItem *item;
 
 item = (MTreeItem *)_structureView->currentItem();
 
 if(MModelView::get()->getViewMode() == MModelView::ViewModeTexture)
 {
 	MMainWindow::get()->onQuadView();
 }
 
 if(item)
 {
  if(item->getData())
  {
   MDocument::get()->setSelectedItem(item->getData());
  }
  else
  {
   MDocument::get()->setSelectedItem(NULL);
  }
 }
 else
 {
  MDocument::get()->setSelectedItem(NULL);
 }

 MItemManipulator::get()->clearVertexStack();
 MMainWindow::get()->updateToolbar();
 MModelView::get()->repaint();
}


void MStructureView::onExpand(void)
{
 MModel *m;
 QPtrList<MModel> models;
 
 MShape *s;
 QPtrList<MShape> shapes;

 // Expand the entire document
 if(!MDocument::get()->getSelectedItem())
 {
  models = MSceneGraph::get()->getModelList();
  for(m = models.first(); m; m = models.next())
  {
   m->setExpanded(true);
  }
 }
 else if(MDocument::get()->getSelectedItem()->getType() == MEnum::Model)
 {
  models = ((MModel *)(MDocument::get()->getSelectedItem()))->getModelList();
  for(m = models.first(); m; m = models.next())
  {
   m->setExpanded(true);
  }
  
  shapes = ((MModel *)(MDocument::get()->getSelectedItem()))->getShapeList();
  for(s = shapes.first(); s; s = shapes.next())
  {
   s->setExpanded(true);
  }
 }

 update();
}


void MStructureView::onCollapse(void)
{
 MModel *m;
 QPtrList<MModel> models;
 
 MShape *s;
 QPtrList<MShape> shapes;

 // Expand the entire document
 if(!MDocument::get()->getSelectedItem())
 {
  models = MSceneGraph::get()->getModelList();
  for(m = models.first(); m; m = models.next())
  {
   m->setExpanded(false);
  }
 }
 else if(MDocument::get()->getSelectedItem()->getType() == MEnum::Model)
 {
  models = ((MModel *)(MDocument::get()->getSelectedItem()))->getModelList();
  for(m = models.first(); m; m = models.next())
  {
   m->setExpanded(false);
  }
  
  shapes = ((MModel *)(MDocument::get()->getSelectedItem()))->getShapeList();
  for(s = shapes.first(); s; s = shapes.next())
  {
   s->setExpanded(false);
  }
 }
 
 update();
}


void MStructureView::expanded(QListViewItem *item)
{
 MTreeItem *titem;
 
 titem = (MTreeItem *)item;
 if(titem->getData())
 {
  titem->getData()->setExpanded(true);
 }
}


void MStructureView::collapsed(QListViewItem *item)
{
 MTreeItem *titem;
 
 titem = (MTreeItem *)item;
 if(titem->getData())
 {
  titem->getData()->setExpanded(false);
 }
}


void MStructureView::itemRenamed(QListViewItem *item, int col, const QString &text)
{
 MTreeItem *titem;
 
 titem = (MTreeItem *)item;
 if(titem->getData())
 {
  MDocument::get()->commit();
  titem->getData()->setName(text);
 }
}



void MStructureView::resizeEvent(QResizeEvent *event)
{
 QSize size = event->size();
 
 _structureView->setGeometry(1, 1, size.width() - 1, size.height() - 1);
}
